import os

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.resizing import AddLinesModule
from pynpoint.processing.timedenoising import CwtWaveletConfiguration, DwtWaveletConfiguration, \
                                              WaveletTimeDenoisingModule, TimeNormalizationModule
from pynpoint.util.tests import create_config, remove_test_data, create_star_data


class TestTimeDenoising:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_star_data(self.test_dir+'images')
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=['images'])

    def test_read_data(self) -> None:

        module = FitsReadingModule(name_in='read',
                                   image_tag='images',
                                   input_dir=self.test_dir+'images',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('read')

        data = self.pipeline.get_data('images')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_wavelet_denoising_cwt_dog(self) -> None:

        cwt_config = CwtWaveletConfiguration(wavelet='dog',
                                             wavelet_order=2,
                                             keep_mean=False,
                                             resolution=0.5)

        assert cwt_config.m_wavelet == 'dog'
        assert cwt_config.m_wavelet_order == 2
        assert not cwt_config.m_keep_mean
        assert cwt_config.m_resolution == 0.5

        module = WaveletTimeDenoisingModule(wavelet_configuration=cwt_config,
                                            name_in='wavelet_cwt_dog',
                                            image_in_tag='images',
                                            image_out_tag='wavelet_cwt_dog',
                                            padding='zero',
                                            median_filter=True,
                                            threshold_function='soft')

        self.pipeline.add_module(module)
        self.pipeline.run_module('wavelet_cwt_dog')

        data = self.pipeline.get_data('wavelet_cwt_dog')
        assert np.sum(data) == pytest.approx(105.1035789572968, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'a') as hdf_file:
            hdf_file['config'].attrs['CPU'] = 4

        self.pipeline.run_module('wavelet_cwt_dog')

        data_multi = self.pipeline.get_data('wavelet_cwt_dog')
        assert data == pytest.approx(data_multi, rel=self.limit, abs=0.)
        assert data.shape == data_multi.shape

    def test_wavelet_denoising_cwt_morlet(self) -> None:

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'a') as hdf_file:
            hdf_file['config'].attrs['CPU'] = 1

        cwt_config = CwtWaveletConfiguration(wavelet='morlet',
                                             wavelet_order=5,
                                             keep_mean=False,
                                             resolution=0.5)

        assert cwt_config.m_wavelet == 'morlet'
        assert cwt_config.m_wavelet_order == 5
        assert not cwt_config.m_keep_mean
        assert cwt_config.m_resolution == 0.5

        module = WaveletTimeDenoisingModule(wavelet_configuration=cwt_config,
                                            name_in='wavelet_cwt_morlet',
                                            image_in_tag='images',
                                            image_out_tag='wavelet_cwt_morlet',
                                            padding='mirror',
                                            median_filter=False,
                                            threshold_function='hard')

        self.pipeline.add_module(module)
        self.pipeline.run_module('wavelet_cwt_morlet')

        data = self.pipeline.get_data('wavelet_cwt_morlet')
        assert np.sum(data) == pytest.approx(104.86262840716438, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

        data = self.pipeline.get_attribute('wavelet_cwt_morlet', 'NFRAMES', static=False)
        assert data[0] == data[1] == 5

    def test_wavelet_denoising_dwt(self) -> None:

        dwt_config = DwtWaveletConfiguration(wavelet='db8')

        assert dwt_config.m_wavelet == 'db8'

        module = WaveletTimeDenoisingModule(wavelet_configuration=dwt_config,
                                            name_in='wavelet_dwt',
                                            image_in_tag='images',
                                            image_out_tag='wavelet_dwt',
                                            padding='zero',
                                            median_filter=True,
                                            threshold_function='soft')

        self.pipeline.add_module(module)
        self.pipeline.run_module('wavelet_dwt')

        data = self.pipeline.get_data('wavelet_dwt')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_time_normalization(self) -> None:

        module = TimeNormalizationModule(name_in='timenorm',
                                         image_in_tag='images',
                                         image_out_tag='timenorm')

        self.pipeline.add_module(module)
        self.pipeline.run_module('timenorm')

        data = self.pipeline.get_data('timenorm')
        assert np.sum(data) == pytest.approx(56.443663773873, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_wavelet_denoising_even_size(self) -> None:

        module = AddLinesModule(name_in='add',
                                image_in_tag='images',
                                image_out_tag='images_even',
                                lines=(1, 0, 1, 0))

        self.pipeline.add_module(module)
        self.pipeline.run_module('add')

        data = self.pipeline.get_data('images_even')
        assert np.sum(data) == pytest.approx(105.54278879805275, rel=self.limit, abs=0.)
        assert data.shape == (10, 12, 12)

        cwt_config = CwtWaveletConfiguration(wavelet='dog',
                                             wavelet_order=2,
                                             keep_mean=False,
                                             resolution=0.5)

        assert cwt_config.m_wavelet == 'dog'
        assert cwt_config.m_wavelet_order == 2
        assert not cwt_config.m_keep_mean
        assert cwt_config.m_resolution == 0.5

        module = WaveletTimeDenoisingModule(wavelet_configuration=cwt_config,
                                            name_in='wavelet_even_1',
                                            image_in_tag='images_even',
                                            image_out_tag='wavelet_even_1',
                                            padding='zero',
                                            median_filter=True,
                                            threshold_function='soft')

        self.pipeline.add_module(module)
        self.pipeline.run_module('wavelet_even_1')

        data = self.pipeline.get_data('wavelet_even_1')
        assert np.sum(data) == pytest.approx(105.1035789572968, rel=self.limit, abs=0.)
        assert data.shape == (10, 12, 12)

        module = WaveletTimeDenoisingModule(wavelet_configuration=cwt_config,
                                            name_in='wavelet_even_2',
                                            image_in_tag='images_even',
                                            image_out_tag='wavelet_even_2',
                                            padding='mirror',
                                            median_filter=True,
                                            threshold_function='soft')

        self.pipeline.add_module(module)
        self.pipeline.run_module('wavelet_even_2')

        data = self.pipeline.get_data('wavelet_even_2')
        assert np.sum(data) == pytest.approx(105.06809820408587, rel=self.limit, abs=0.)
        assert data.shape == (10, 12, 12)

        data = self.pipeline.get_attribute('images', 'NFRAMES', static=False)
        assert data == pytest.approx([5, 5], rel=self.limit, abs=0.)

        data = self.pipeline.get_attribute('wavelet_even_1', 'NFRAMES', static=False)
        assert data == pytest.approx([5, 5], rel=self.limit, abs=0.)

        data = self.pipeline.get_attribute('wavelet_even_2', 'NFRAMES', static=False)
        assert data == pytest.approx([5, 5], rel=self.limit, abs=0.)



"""
Pipeline modules for photometric and astrometric measurements.
"""

import os
import time
import warnings

from typing import Any, List, Optional, Tuple, Union
from multiprocessing import Pool

import numpy as np
import emcee

from typeguard import typechecked
from scipy.optimize import minimize
from sklearn.decomposition import PCA
from photutils.aperture import CircularAperture

from pynpoint.core.processing import ProcessingModule
from pynpoint.util.apply_func import photometry
from pynpoint.util.analysis import fake_planet, merit_function, false_alarm, pixel_variance
from pynpoint.util.image import create_mask, polar_to_cartesian, cartesian_to_polar, \
                                center_subpixel, rotate_coordinates
from pynpoint.util.mcmc import lnprob
from pynpoint.util.module import progress, memory_frames
from pynpoint.util.psf import pca_psf_subtraction
from pynpoint.util.residuals import combine_residuals


class FakePlanetModule(ProcessingModule):
    """
    Pipeline module to inject a positive or negative artificial planet
    into a stack of images.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 psf_in_tag: str,
                 image_out_tag: str,
                 position: Tuple[float, float],
                 magnitude: float,
                 psf_scaling: float = 1.,
                 interpolation: str = 'spline') -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry with images that are read as
            input.
        psf_in_tag : str
            Tag of the database entry that contains the reference PSF
            that is used as fake planet. Can be either a single image
            (2D) or a cube (3D) with the dimensions equal to
            ``image_in_tag``.
        image_out_tag : str
            Tag of the database entry with images that are written as
            output.
        position : tuple(float, float)
            Angular separation (arcsec) and position angle (deg) of
            the fake planet. Angle is measured in counterclockwise
            direction with respect to the upward direction (i.e.,
            East of North).
        magnitude : float
            Magnitude of the fake planet with respect to the star.
        psf_scaling : float
            Additional scaling factor of the planet flux (e.g., to
            correct for a neutral density filter). A negative value
            will inject a negative planet signal.
        interpolation : str
            Type of interpolation that is used for shifting the
            images (spline, bilinear, or fft).

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)

        if psf_in_tag == image_in_tag:
            self.m_psf_in_port = self.m_image_in_port
        else:
            self.m_psf_in_port = self.add_input_port(psf_in_tag)

        self.m_image_out_port = self.add_output_port(image_out_tag)

        self.m_position = position
        self.m_magnitude = magnitude
        self.m_psf_scaling = psf_scaling
        self.m_interpolation = interpolation

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Shifts the PSF template to the
        location of the fake planet with an additional correction
        for the parallactic angle and an optional flux scaling. The
        stack of images with the injected planet signal is stored.

        Returns
        -------
        NoneType
            None
        """

        print('Input parameters:')
        print(f'   - Magnitude = {self.m_magnitude:.2f}')
        print(f'   - PSF scaling = {self.m_psf_scaling}')
        print(f'   - Separation (arcsec) = {self.m_position[0]:.2f}')
        print(f'   - Position angle (deg) = {self.m_position[1]:.2f}')

        memory = self._m_config_port.get_attribute('MEMORY')
        parang = self.m_image_in_port.get_attribute('PARANG')
        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')

        self.m_position = (self.m_position[0]/pixscale, self.m_position[1])

        im_shape = self.m_image_in_port.get_shape()
        psf_shape = self.m_psf_in_port.get_shape()

        if psf_shape[0] != 1 and psf_shape[0] != im_shape[0]:
            raise ValueError('The number of frames in psf_in_tag does not match with the number '
                             'of frames in image_in_tag. The DerotateAndStackModule can be '
                             'used to average the PSF frames (without derotating) before applying '
                             'the FakePlanetModule.')

        if psf_shape[-2:] != im_shape[-2:]:
            raise ValueError(f'The images in \'{self.m_image_in_port.tag}\' should have the same '
                             f'dimensions as the images images in \'{self.m_psf_in_port.tag}\'.')

        frames = memory_frames(memory, im_shape[0])

        start_time = time.time()
        for j, _ in enumerate(frames[:-1]):
            progress(j, len(frames[:-1]), 'Injecting artificial planets...', start_time)

            images = self.m_image_in_port[frames[j]:frames[j+1]]
            angles = parang[frames[j]:frames[j+1]]

            if psf_shape[0] == 1:
                psf = self.m_psf_in_port.get_all()
            else:
                psf = self.m_psf_in_port[frames[j]:frames[j+1]]

            im_fake = fake_planet(images=images,
                                  psf=psf,
                                  parang=angles,
                                  position=self.m_position,
                                  magnitude=self.m_magnitude,
                                  psf_scaling=self.m_psf_scaling,
                                  interpolation='spline')

            self.m_image_out_port.append(im_fake, data_dim=3)

        history = f'(sep, angle, mag) = ({self.m_position[0]*pixscale:.2f}, ' \
                  f'{self.m_position[1]:.2f}, {self.m_magnitude:.2f})'

        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.add_history('FakePlanetModule', history)
        self.m_image_out_port.close_port()


class SimplexMinimizationModule(ProcessingModule):
    """
    Pipeline module to retrieve the contrast and position of a planet
    by injecting negative artificial planets and using a downhill
    simplex method. The module supports both ADI and RDI.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 psf_in_tag: str,
                 res_out_tag: str,
                 flux_position_tag: str,
                 position: Tuple[float, float],
                 magnitude: float,
                 psf_scaling: float = -1.,
                 merit: str = 'gaussian',
                 aperture: float = 0.1,
                 sigma: float = 0.0,
                 tolerance: float = 0.1,
                 pca_number: Union[int, range, List[int]] = 10,
                 cent_size: Optional[float] = None,
                 edge_size: Optional[float] = None,
                 extra_rot: float = 0.,
                 residuals: str = 'median',
                 reference_in_tag: Optional[str] = None,
                 offset: Optional[float] = None) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry with the science images that are
            read as input.
        psf_in_tag : str
            Tag of the database entry with the reference PSF that is
            used as fake planet. Can be either a single image or a
            stack of images equal in size to ``image_in_tag``.
        res_out_tag : str
            Tag of the database entry with the image residuals that
            are written as output. The residuals are stored for each
            step of the minimization. The last image contains the
            best-fit residuals.
        flux_position_tag : str
            Tag of the database entry with the flux and position
            results that are written as output. Each step of the
            minimization stores the x position (pixels), y position
            (pixels), separation (arcsec), angle (deg), contrast
            (mag), and the chi-square value. The last row contains
            the best-fit results.
        position : tuple(float, float)
            Approximate position of the planet (x, y), provided with
            subpixel precision (i.e. as floats). The figure of merit
            is calculated within an aperture of radius ``aperture``
            centered at the rounded (i.e. integers) coordinates of
            ``position``. When setting, ``offset=0.``, the ``position``
            is used as fixed position of the planet while only
            retrieving the contrast.
        magnitude : float
            Approximate magnitude of the planet relative to the star.
        psf_scaling : float
            Additional scaling factor of the planet flux (e.g., to
            correct for a neutral density filter). Should be a negative
            value in order to inject negative fake planets.
        merit : str
            Figure of merit for the minimization ('hessian', 'gaussian',
            or 'poisson'). Either the determinant of the Hessian matrix
            is minimized ('hessian') or the flux of each pixel
            ('gaussian' or 'poisson'). For the latter case, the
            estimated noise is assumed to follow a Poisson (see Wertz et
            al. 2017) or Gaussian distribution (see Stolker et al. 2020).
        aperture : float
            Aperture radius (arcsec) at the position specified at
            ``position``.
        sigma : float
            Standard deviation (arcsec) of the Gaussian kernel that is
            used to smooth the images before the figure of merit is
            calculated (in order to reduce small pixel-to-pixel
            variations).
        tolerance : float
            Absolute error on the input parameters, position (pixels)
            and contrast (mag), that is used as acceptance level for
            convergence. Note that only a single value can be specified
            which is used for both the position and flux so
            ``tolerance=0.1`` corresponds to a precision of 0.1 mag and
            0.1 pix. The tolerance on the output (i.e., the chi-square
            value) is set to ``np.inf`` such that the condition is
            always met.
        pca_number : int, range, list(int, )
            Number of principal components (PCs) used for the PSF
            subtraction. Can be either a single value, or a range or
            list of values. In the latter case, the ``res_out_tag`` and
            ```flux_position_tag``` contain a 3 digit number with the
             number of PCs.
        cent_size : float, None
            Radius of the central mask (arcsec). No mask is used when
            set to ``None``. The mask is applied after the artificial
            planet is injected.
        edge_size : float, None
            Outer radius (arcsec) beyond which pixels are masked. No
            outer mask is used when set to ``None``. The radius will be
            set to half the image size if the argument is larger than
            half the image size. The mask is applied after the
            artificial planet is injected.
        extra_rot : float
            Additional rotation angle of the images in clockwise
            direction (deg).
        residuals : str
            Method for combining the residuals ('mean', 'median',
            'weighted', or 'clipped').
        reference_in_tag : str, None
            Tag of the database entry with the reference images that
            are read as input. The data of the ``image_in_tag`` itself
            is used as reference data for the PSF subtraction if set to
            ``None``. Note that the mean is not subtracted from the
            data of ``image_in_tag`` and ``reference_in_tag`` in case
            the ``reference_in_tag`` is used, to allow for flux and
            position measurements in the context of RDI.
        offset : float, None
            Offset (pixels) by which the injected negative PSF may
            deviate from ``position``. The constraint on the position
            is not applied if set to None. Only the contrast is
            optimized and the position if fixed to ``position`` if
            ``offset=0``.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)

        if psf_in_tag == image_in_tag:
            self.m_psf_in_port = self.m_image_in_port
        else:
            self.m_psf_in_port = self.add_input_port(psf_in_tag)

        if reference_in_tag is None:
            self.m_reference_in_port = None
        else:
            self.m_reference_in_port = self.add_input_port(reference_in_tag)

        self.m_res_out_port = []
        self.m_flux_pos_port = []

        if isinstance(pca_number, int):
            self.m_res_out_port.append(self.add_output_port(res_out_tag))
            self.m_flux_pos_port.append(self.add_output_port(flux_position_tag))

        else:
            for item in pca_number:
                self.m_res_out_port.append(self.add_output_port(res_out_tag+f'{item:03d}'))
                self.m_flux_pos_port.append(self.add_output_port(flux_position_tag+f'{item:03d}'))

        self.m_position = position
        self.m_magnitude = magnitude
        self.m_psf_scaling = psf_scaling
        self.m_merit = merit
        self.m_aperture = aperture
        self.m_sigma = sigma
        self.m_tolerance = tolerance
        self.m_cent_size = cent_size
        self.m_edge_size = edge_size
        self.m_extra_rot = extra_rot
        self.m_residuals = residuals
        self.m_offset = offset

        if isinstance(pca_number, int):
            self.m_pca_number = [pca_number]

        else:
            self.m_pca_number = pca_number

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. The position and contrast of a planet
        is measured by injecting negative copies of the PSF template
        and applying a downhill simplex method (Nelder-Mead) for
        minimization of a figure of merit at the planet location.

        Returns
        -------
        NoneType
            None
        """

        print('Input parameters:')
        print(f'   - Number of principal components = {self.m_pca_number}')
        print(f'   - Figure of merit = {self.m_merit}')
        print(f'   - Residuals type = {self.m_residuals}')
        print(f'   - Absolute tolerance (pixels/mag) = {self.m_tolerance}')
        print(f'   - Maximum offset = {self.m_offset}')
        print(f'   - Guessed position (x, y) = ({self.m_position[0]:.2f}, '
              f'{self.m_position[1]:.2f})')

        parang = self.m_image_in_port.get_attribute('PARANG')
        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')

        aperture = (round(self.m_position[1]), round(self.m_position[0]), self.m_aperture/pixscale)
        print(f'   - Aperture position (x, y) = ({aperture[1]}, {aperture[0]})')
        print(f'   - Aperture radius (pixels) = {int(aperture[2])}')

        self.m_sigma /= pixscale

        if self.m_cent_size is not None:
            self.m_cent_size /= pixscale
            print(f'   - Inner mask radius (pixels) = {int(self.m_cent_size)}')

        if self.m_edge_size is not None:
            self.m_edge_size /= pixscale
            print(f'   - Outer mask radius (pixels) = {int(self.m_edge_size)}')

        psf = self.m_psf_in_port.get_all()
        images = self.m_image_in_port.get_all()

        if psf.shape[0] != 1 and psf.shape[0] != images.shape[0]:
            raise ValueError('The number of frames in psf_in_tag does not match with the number '
                             'of frames in image_in_tag. The DerotateAndStackModule can be '
                             'used to average the PSF frames (without derotating) before applying '
                             'the SimplexMinimizationModule.')

        center = center_subpixel(psf)
        print(f'Image center (y, x) = {center}')

        # Rotate the initial position, (y, x), by the extra rotation angle to (y_rot, x_rot)
        pos_init = rotate_coordinates(center,
                                      (self.m_position[1], self.m_position[0]),
                                      self.m_extra_rot)

        if self.m_reference_in_port is not None and self.m_merit != 'poisson':
            raise NotImplementedError('The reference_in_tag can only be used in combination with '
                                      'the \'poisson\' figure of merit.')

        @typechecked
        def _objective(arg: np.ndarray,
                       count: int,
                       n_components: int,
                       sklearn_pca: Optional[PCA],
                       var_noise: Optional[float]) -> float:

            # Extract the contrast, y position, and x position from the argument tuple
            mag = arg[0]

            if self.m_offset is None or self.m_offset > 0.:
                pos_y = arg[1]
                pos_x = arg[2]

            else:
                pos_y = pos_init[0]
                pos_x = pos_init[1]

            # Calculate the absolute offset (pixels) with respect to the initial guess
            pos_offset = np.sqrt((pos_x-pos_init[1])**2 + (pos_y-pos_init[0])**2)

            if self.m_offset is not None and pos_offset > self.m_offset:
                # Return chi-square = inf if the offset needs to be tested and is too large
                return np.inf

            # Convert the cartesian position to a separation and position angle
            sep_ang = cartesian_to_polar(center, pos_y, pos_x)

            # Inject the negative artifical planet at the position and contrast that is tested
            fake = fake_planet(images=images,
                               psf=psf,
                               parang=parang,
                               position=(sep_ang[0], sep_ang[1]),
                               magnitude=mag,
                               psf_scaling=self.m_psf_scaling)

            # Create a mask
            mask = create_mask(fake.shape[-2:], (self.m_cent_size, self.m_edge_size))

            if self.m_reference_in_port is None:
                # PSF subtraction with the science data as reference data (ADI)
                im_res_rot, im_res_derot = pca_psf_subtraction(images=fake*mask,
                                                               angles=-1.*parang+self.m_extra_rot,
                                                               pca_number=n_components,
                                                               pca_sklearn=sklearn_pca,
                                                               im_shape=None,
                                                               indices=None)

            else:
                # PSF subtraction with separate reference data (RDI)
                im_reshape = np.reshape(fake*mask, (im_shape[0], im_shape[1]*im_shape[2]))

                im_res_rot, im_res_derot = pca_psf_subtraction(images=im_reshape,
                                                               angles=-1.*parang+self.m_extra_rot,
                                                               pca_number=n_components,
                                                               pca_sklearn=sklearn_pca,
                                                               im_shape=im_shape,
                                                               indices=None)

            # Collapse the residuals of the PSF subtraction
            res_stack = combine_residuals(method=self.m_residuals,
                                          res_rot=im_res_derot,
                                          residuals=im_res_rot,
                                          angles=parang)

            # Appedn the collapsed residuals to the output port
            self.m_res_out_port[count].append(res_stack, data_dim=3)

            # Calculate the chi-square for the tested position and contrast
            chi_sq = merit_function(residuals=res_stack[0, ],
                                    merit=self.m_merit,
                                    aperture=aperture,
                                    sigma=self.m_sigma,
                                    var_noise=var_noise)

            # Apply the extra rotation to the y and x position
            # The returned position is given as (y, x)
            position = rotate_coordinates(center, (pos_y, pos_x), -self.m_extra_rot)

            # Create and array with the x position, y position, separation (arcsec), position
            # angle (deg), contrast (mag), and chi-square
            res = np.asarray([position[1],
                              position[0],
                              sep_ang[0]*pixscale,
                              (sep_ang[1]-self.m_extra_rot) % 360.,
                              mag,
                              chi_sq])

            # Append the results to the output port
            self.m_flux_pos_port[count].append(res, data_dim=2)

            print(f'\rSimplex minimization... {n_components} PC - chi^2 = {chi_sq:.2e}', end='')

            return chi_sq

        for i, n_components in enumerate(self.m_pca_number):
            print(f'\rSimplex minimization... {n_components} PC ', end='')

            if self.m_reference_in_port is None:
                sklearn_pca = None

            else:
                ref_data = self.m_reference_in_port.get_all()

                im_shape = images.shape
                ref_shape = ref_data.shape

                if ref_shape[1:] != im_shape[1:]:
                    raise ValueError('The image size of the science data and the reference data '
                                     'should be identical.')

                # reshape reference data and select the unmasked pixels
                ref_reshape = ref_data.reshape(ref_shape[0], ref_shape[1]*ref_shape[2])

                mean_ref = np.mean(ref_reshape, axis=0)
                ref_reshape -= mean_ref

                # create the PCA basis
                sklearn_pca = PCA(n_components=n_components, svd_solver='arpack')
                sklearn_pca.fit(ref_reshape)

                # add mean of reference array as 1st PC and orthogonalize it to the PCA basis
                mean_ref_reshape = mean_ref.reshape((1, mean_ref.shape[0]))

                q_ortho, _ = np.linalg.qr(np.vstack((mean_ref_reshape,
                                                     sklearn_pca.components_[:-1, ])).T)

                sklearn_pca.components_ = q_ortho.T

            if self.m_merit == 'poisson':
                var_noise = None

            elif self.m_merit in ['gaussian', 'hessian']:
                var_noise = pixel_variance(var_type=self.m_merit,
                                           images=images,
                                           parang=parang,
                                           cent_size=self.m_cent_size,
                                           edge_size=self.m_edge_size,
                                           pca_number=n_components,
                                           residuals=self.m_residuals,
                                           aperture=aperture,
                                           sigma=self.m_sigma)

            if self.m_offset == 0.:
                x0_minimize = np.array([self.m_magnitude])
            else:
                x0_minimize = np.array([self.m_magnitude, pos_init[0], pos_init[1]])

            min_result = minimize(fun=_objective,
                                  x0=x0_minimize,
                                  args=(i, n_components, sklearn_pca, var_noise),
                                  method='Nelder-Mead',
                                  tol=None,
                                  options={'xatol': self.m_tolerance, 'fatol': float('inf')})

            print(' [DONE]')

            if self.m_offset == 0.:
                pos_x = pos_init[1]
                pos_y = pos_init[0]

            else:
                pos_x = min_result.x[2]
                pos_y = min_result.x[1]

            pos_rot_yx = rotate_coordinates(center, (pos_y, pos_x), -self.m_extra_rot)

            sep_ang = cartesian_to_polar(center, pos_rot_yx[0], pos_rot_yx[1])

            print('Best-fit parameters:')
            print(f'   - Position (x, y) = ({pos_rot_yx[1]:.2f}, {pos_rot_yx[0]:.2f})')
            print(f'   - Separation (mas) = {sep_ang[0]*pixscale*1e3:.2f}')
            print(f'   - Position angle (deg) = {sep_ang[1]:.2f}')
            print(f'   - Contrast (mag) = {min_result.x[0]:.2f}')

        history = f'merit = {self.m_merit}'

        for item in self.m_flux_pos_port:
            item.copy_attributes(self.m_image_in_port)
            item.add_history('SimplexMinimizationModule', history)

        for item in self.m_res_out_port:
            item.copy_attributes(self.m_image_in_port)
            item.add_history('SimplexMinimizationModule', history)

        self.m_res_out_port[0].close_port()


class FalsePositiveModule(ProcessingModule):
    """
    Pipeline module to calculate the signal-to-noise ratio (SNR) and false positive fraction (FPF)
    at a specified location in an image by using the Student's t-test (Mawet et al. 2014).
    Optionally, the SNR can be optimized with the aperture position as free parameter.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 snr_out_tag: str,
                 position: Tuple[float, float],
                 aperture: float = 0.1,
                 ignore: bool = False,
                 optimize: bool = False,
                 **kwargs: Any) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry with the images that are read as input. The SNR/FPF is
            calculated for each image in the dataset.
        snr_out_tag : str
            Tag of the database entry that is written as output. The output format is: (x position
            (pix), y position (pix), separation (arcsec), position angle (deg), SNR, FPF). The
            position angle is measured in counterclockwise direction with respect to the upward
            direction (i.e., East of North).
        position : tuple(float, float)
            The x and y position (pix) where the SNR and FPF is calculated. Note that the bottom
            left of the image is defined as (-0.5, -0.5) so there is a -1.0 offset with respect
            to the DS9 coordinate system. Aperture photometry corrects for the partial inclusion
            of pixels at the boundary.
        aperture : float
            Aperture radius (arcsec).
        ignore : bool
            Ignore the two neighboring apertures that may contain self-subtraction from the planet.
        optimize : bool
            Optimize the SNR. The aperture position is stored in the `snr_out_tag`. The size of the
            aperture is kept fixed.

        Keyword arguments
        -----------------
        tolerance : float
            The absolute tolerance on the position for the optimization to end. Default is set
            to 0.01 (pix).
        offset : float, None
            Offset (pix) by which the aperture may deviate from ``position`` when
            ``optimize=True`` (default: None).

        Returns
        -------
        NoneType
            None
        """

        if 'tolerance' in kwargs:
            self.m_tolerance = kwargs['tolerance']
        else:
            self.m_tolerance = 1e-2

        if 'offset' in kwargs:
            self.m_offset = kwargs['offset']
        else:
            self.m_offset = None

        if 'bounds' in kwargs:
            warnings.warn('The \'bounds\' keyword argument has been deprecated. Please use '
                          '\'offset\' instead (e.g. offset=3.0).', DeprecationWarning)

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_snr_out_port = self.add_output_port(snr_out_tag)

        self.m_position = position
        self.m_aperture = aperture
        self.m_ignore = ignore
        self.m_optimize = optimize

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Calculates the SNR and FPF for a specified position in a post-
        processed image with the Student's t-test (Mawet et al. 2014). This approach assumes
        Gaussian noise but accounts for small sample statistics.

        Returns
        -------
        NoneType
            None
        """

        @typechecked
        def _snr_optimize(arg: np.ndarray) -> float:

            pos_x, pos_y = arg

            pos_offset = np.sqrt((pos_x-self.m_position[0])**2 + (pos_y-self.m_position[1])**2)

            if self.m_offset is not None:
                if pos_offset > self.m_offset:
                    snr = 0.

                else:
                    snr = None

            if self.m_offset is None or snr is None:
                _, _, snr, _ = false_alarm(image=image,
                                           x_pos=pos_x,
                                           y_pos=pos_y,
                                           size=self.m_aperture,
                                           ignore=self.m_ignore)

            return -1.*snr

        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')
        self.m_aperture /= pixscale

        print('Input parameters:')
        print(f'   - Aperture position = {self.m_position}')
        print(f'   - Aperture radius (pixels) = {self.m_aperture:.2f}')
        print(f'   - Optimize aperture position = {self.m_optimize}')
        print(f'   - Ignore neighboring apertures = {self.m_ignore}')
        print(f'   - Minimization tolerance = {self.m_tolerance}')

        nimages = self.m_image_in_port.get_shape()[0]

        print('Calculating the S/N and FPF...')

        for j in range(nimages):
            image = self.m_image_in_port[j, ]
            center = center_subpixel(image)

            if self.m_optimize:
                result = minimize(fun=_snr_optimize,
                                  x0=np.array([self.m_position[0], self.m_position[1]]),
                                  method='Nelder-Mead',
                                  tol=None,
                                  options={'xatol': self.m_tolerance, 'fatol': float('inf')})

                _, _, snr, fpf = false_alarm(image=image,
                                             x_pos=result.x[0],
                                             y_pos=result.x[1],
                                             size=self.m_aperture,
                                             ignore=self.m_ignore)

                x_pos, y_pos = result.x[0], result.x[1]

            else:
                _, _, snr, fpf = false_alarm(image=image,
                                             x_pos=self.m_position[0],
                                             y_pos=self.m_position[1],
                                             size=self.m_aperture,
                                             ignore=self.m_ignore)

                x_pos, y_pos = self.m_position[0], self.m_position[1]

            print(f'Image {j+1:03d}/{nimages} -> (x, y) = ({x_pos:.2f}, {y_pos:.2f}), '
                  f'S/N = {snr:.2f}, FPF = {fpf:.2e}')

            sep_ang = cartesian_to_polar(center, y_pos, x_pos)
            result = np.column_stack((x_pos, y_pos, sep_ang[0]*pixscale, sep_ang[1], snr, fpf))

            self.m_snr_out_port.append(result, data_dim=2)

        history = f'aperture (arcsec) = {self.m_aperture*pixscale:.2f}'
        self.m_snr_out_port.copy_attributes(self.m_image_in_port)
        self.m_snr_out_port.add_history('FalsePositiveModule', history)
        self.m_snr_out_port.close_port()


class MCMCsamplingModule(ProcessingModule):
    """
    Pipeline module to measure the separation, position angle, and
    contrast of a planet with injection of negative artificial planets
    and sampling of the posterior distribution with ``emcee``, an
    affine invariant Markov chain Monte Carlo (MCMC) ensemble sampler.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 psf_in_tag: str,
                 chain_out_tag: str,
                 param: Tuple[float, float, float],
                 bounds: Tuple[Tuple[float, float], Tuple[float, float], Tuple[float, float]],
                 nwalkers: int = 100,
                 nsteps: int = 200,
                 psf_scaling: float = -1.,
                 pca_number: int = 20,
                 aperture: Union[float, Tuple[int, int, float]] = 0.1,
                 mask: Optional[Union[Tuple[float, float], Tuple[None, float],
                                      Tuple[float, None], Tuple[None, None]]] = None,
                 extra_rot: float = 0.,
                 merit: str = 'gaussian',
                 residuals: str = 'median',
                 resume: bool = False,
                 **kwargs: Union[float, Tuple[float, float, float]]) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Database tag with the science images.
        psf_in_tag : str
            Database tag with the reference PSF that is used as
            artificial planet. The dataset can be either a single
            image, or a stack of images with the dimensions equal
            to ``image_in_tag``.
        chain_out_tag : str
            Database tag were the posterior samples will be stored.
            The shape of the array is ``(nsteps, nwalkers, 3)``. The
            mean acceptance fraction and the integrated autocorrelation
            time are stored as attributes.
        param : tuple(float, float, float)
            The approximate separation (arcsec), angle (deg), and
            contrast (mag), for example obtained with the
            :class:`~pynpoint.processing.fluxposition.SimplexMinimizationModule`.
            The angle is measured in counterclockwise direction with
            respect to the upward direction (i.e., East of North). The
            separation and angle are also used as (fixed) position for the
            aperture if ``aperture`` contains a float (i.e. the radius).
        bounds : tuple(tuple(float, float), tuple(float, float), tuple(float, float))
            The prior boundaries for the separation (arcsec), angle
            (deg), and contrast (mag). Each set of boundaries is
            specified as a tuple.
        nwalkers : int
            Number of walkers.
        nsteps : int
            Number of steps per walker.
        psf_scaling : float
            Additional scaling factor of the planet flux (e.g. to
            correct for a neutral density filter or difference in
            exposure time). The value should be negative in order
            to inject negative fake planets.
        pca_number : int
            Number of principal components used for the PSF
            subtraction.
        aperture : float, tuple(int, int, float)
            Either the aperture radius (arcsec) at the position of
            ``param`` or tuple with the position and aperture radius
            (arcsec) as ``(pos_x, pos_y, radius)``.
        mask : tuple(float, float), None
            Inner and outer mask radius (arcsec) for the PSF
            subtraction. Both elements of the tuple can be set to
            ``None``. Masked pixels are excluded from the PCA
            computation, resulting in a smaller runtime. Masking is
            done after the artificial planet is injected.
        extra_rot : float
            Additional rotation angle of the images (deg).
        merit : str
            Figure of merit for the minimization ('hessian',
            'gaussian', or 'poisson'). Either the determinant of the
            Hessian matrix is minimized ('hessian') or the flux of each
            pixel ('gaussian' or 'poisson'). For the latter case, the
            estimate noise is assumed to follow a Poisson (see Wertz et
            al. 2017) or Gaussian distribution (see Wertz et al. 2017
            and Stolker et al. 2020).
        residuals : str
            Method used for combining the residuals ('mean', 'median',
            'weighted', or 'clipped').
        resume : bool
            Resume from the last state of the chain that was stored by
            the backend of ``emcee``. Set to ``True`` for continuing
            with the samples from a previous run, for example when it
            was interrupted or to create more steps for the walkers.
            The backend data of ``emcee`` is stored with the tag
            ``[chain_out_tag]_backend`` in the HDF5 database.

        Keyword arguments
        -----------------
        sigma : tuple(float, float, float)
            The standard deviations that randomly initializes the
            start positions of the walkers in a small ball around
            the a priori preferred position. The tuple should contain
            a value for the separation (arcsec), position angle (deg),
            and contrast (mag). The default is set to
            ``(1e-5, 1e-3, 1e-3)``.

        Returns
        -------
        NoneType
            None
        """

        if 'sigma' in kwargs:
            self.m_sigma = kwargs['sigma']
        else:
            self.m_sigma = (1e-5, 1e-3, 1e-3)

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)

        if psf_in_tag == image_in_tag:
            self.m_psf_in_port = self.m_image_in_port
        else:
            self.m_psf_in_port = self.add_input_port(psf_in_tag)

        self.m_chain_out_port = self.add_output_port(chain_out_tag)

        self.m_param = param
        self.m_bounds = bounds
        self.m_nwalkers = nwalkers
        self.m_nsteps = nsteps
        self.m_psf_scaling = psf_scaling
        self.m_pca_number = pca_number
        self.m_aperture = aperture
        self.m_extra_rot = extra_rot
        self.m_merit = merit
        self.m_residuals = residuals
        self.m_resume = resume

        if mask is None:
            self.m_mask = (None, None)
        else:
            self.m_mask = mask

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. The posterior distributions of the
        separation, position angle, and flux contrast are sampled with
        the affine invariant Markov chain Monte Carlo (MCMC) ensemble
        sampler ``emcee``. At each step, a negative copy of the PSF
        template is injected and the likelihood function is evaluated
        at the approximate position of the planet.

        Returns
        -------
        NoneType
            None
        """

        print('Input parameters:')
        print(f'   - Number of principal components: {self.m_pca_number}')
        print(f'   - Figure of merit: {self.m_merit}')

        ndim = 3

        cpu = self._m_config_port.get_attribute('CPU')
        work_place = self._m_config_port.get_attribute('WORKING_PLACE')
        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')
        parang = self.m_image_in_port.get_attribute('PARANG')

        images = self.m_image_in_port.get_all()
        psf = self.m_psf_in_port.get_all()

        im_shape = self.m_image_in_port.get_shape()[-2:]

        self.m_image_in_port.close_port()
        self.m_psf_in_port.close_port()

        if psf.shape[0] != 1 and psf.shape[0] != images.shape[0]:
            raise ValueError('The number of frames in psf_in_tag does not match with the number of '
                             'frames in image_in_tag. The DerotateAndStackModule can be used to '
                             'average the PSF frames (without derotating) before applying the '
                             'MCMCsamplingModule.')

        if self.m_mask[0] is not None:
            self.m_mask = (self.m_mask[0]/pixscale, self.m_mask[1])

        if self.m_mask[1] is not None:
            self.m_mask = (self.m_mask[0], self.m_mask[1]/pixscale)

        # create the mask and get the unmasked image indices
        mask = create_mask(im_shape[-2:], self.m_mask)
        indices = np.where(mask.reshape(-1) != 0.)[0]

        if isinstance(self.m_aperture, float):
            yx_pos = polar_to_cartesian(images, self.m_param[0]/pixscale, self.m_param[1])
            aperture = (round(yx_pos[0]), round(yx_pos[1]), self.m_aperture/pixscale)

        elif isinstance(self.m_aperture, tuple):
            aperture = (self.m_aperture[1], self.m_aperture[0], self.m_aperture[2]/pixscale)

        print(f'   - Aperture position (x, y): ({aperture[1]}, {aperture[0]})')
        print(f'   - Aperture radius (pixels): {int(aperture[2])}')

        if self.m_merit == 'poisson':
            var_noise = None

        elif self.m_merit in ['gaussian', 'hessian']:
            var_noise = pixel_variance(var_type=self.m_merit,
                                       images=images,
                                       parang=parang,
                                       cent_size=self.m_mask[0],
                                       edge_size=self.m_mask[1],
                                       pca_number=self.m_pca_number,
                                       residuals=self.m_residuals,
                                       aperture=aperture,
                                       sigma=0.)

            if self.m_merit == 'gaussian':
                print(f'Gaussian standard deviation (counts): {np.sqrt(var_noise):.2e}')
            elif self.m_merit == 'hessian':
                print(f'Hessian standard deviation: {np.sqrt(var_noise):.2e}')

        initial = np.zeros((self.m_nwalkers, ndim))

        initial[:, 0] = self.m_param[0] + np.random.normal(0, self.m_sigma[0], self.m_nwalkers)
        initial[:, 1] = self.m_param[1] + np.random.normal(0, self.m_sigma[1], self.m_nwalkers)
        initial[:, 2] = self.m_param[2] + np.random.normal(0, self.m_sigma[2], self.m_nwalkers)

        backend = emcee.backends.HDFBackend(
            os.path.join(work_place, 'PynPoint_database.hdf5'),
            name=self.m_chain_out_port.tag+'_backend', read_only=False)

        if not self.m_resume:
            print('Reset backend of emcee...', end='', flush=True)
            backend.reset(self.m_nwalkers, ndim)
            print(' [DONE]')

        print('Sampling the posterior distributions with MCMC...')

        with Pool(processes=cpu) as pool:
            sampler = emcee.EnsembleSampler(self.m_nwalkers,
                                            ndim,
                                            lnprob,
                                            pool=pool,
                                            args=([self.m_bounds,
                                                   images,
                                                   psf,
                                                   mask,
                                                   parang,
                                                   self.m_psf_scaling,
                                                   pixscale,
                                                   self.m_pca_number,
                                                   self.m_extra_rot,
                                                   aperture,
                                                   indices,
                                                   self.m_merit,
                                                   self.m_residuals,
                                                   var_noise]),
                                            backend=backend)

            sampler.run_mcmc(initial, self.m_nsteps, progress=True)

        samples = sampler.get_chain()

        self.m_image_in_port._check_status_and_activate()
        self.m_chain_out_port._check_status_and_activate()

        self.m_chain_out_port.set_all(samples)
        print(f'Number of samples stored: {samples.shape[0]*samples.shape[1]}')

        burnin = int(0.2*samples.shape[0])
        samples = samples[burnin:, :, :].reshape((-1, ndim))

        sep_percen = np.percentile(samples[:, 0], [16., 50., 84.])
        ang_percen = np.percentile(samples[:, 1], [16., 50., 84.])
        mag_percen = np.percentile(samples[:, 2], [16., 50., 84.])

        print('Median and uncertainties (20% removed as burnin):')

        print(f'Separation (mas) = {1e3*sep_percen[1]:.2f} '
              f'(-{1e3*sep_percen[1]-1e3*sep_percen[0]:.2f} '
              f'+{1e3*sep_percen[2]-1e3*sep_percen[1]:.2f})')

        print(f'Position angle (deg) = {ang_percen[1]:.2f} '
              f'(-{ang_percen[1]-ang_percen[0]:.2f} '
              f'+{ang_percen[2]-ang_percen[1]:.2f})')

        print(f'Contrast (mag) = {mag_percen[1]:.2f} '
              f'(-{mag_percen[1]-mag_percen[0]:.2f} '
              f'+{mag_percen[2]-mag_percen[1]:.2f})')

        history = f'walkers = {self.m_nwalkers}, steps = {self.m_nsteps}'
        self.m_chain_out_port.copy_attributes(self.m_image_in_port)
        self.m_chain_out_port.add_history('MCMCsamplingModule', history)

        mean_accept = np.mean(sampler.acceptance_fraction)
        print(f'Mean acceptance fraction: {mean_accept:.3f}')
        self.m_chain_out_port.add_attribute('ACCEPTANCE', mean_accept, static=True)

        try:
            autocorr = emcee.autocorr.integrated_time(sampler.get_chain())
            print(f'Integrated autocorrelation time = {autocorr}')

        except emcee.autocorr.AutocorrError:
            autocorr = [np.nan, np.nan, np.nan]
            print('The chain is too short to reliably estimate the autocorrelation time. [WARNING]')

        self.m_chain_out_port.add_attribute('AUTOCORR_0', autocorr[0], static=True)
        self.m_chain_out_port.add_attribute('AUTOCORR_1', autocorr[1], static=True)
        self.m_chain_out_port.add_attribute('AUTOCORR_2', autocorr[2], static=True)

        self.m_chain_out_port.close_port()


class AperturePhotometryModule(ProcessingModule):
    """
    Pipeline module for calculating the counts within a circular area.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 phot_out_tag: str,
                 radius: float = 0.1,
                 position: Optional[Tuple[float, float]] = None) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that is read as input.
        phot_out_tag : str
            Tag of the database entry with the photometry values that are written as output.
        radius : float
            Radius (arcsec) of the circular aperture.
        position : tuple(float, float), None
            Center position (pix) of the aperture, (x, y), with subpixel precision. The center of
            the image will be used if set to None. Python indexing starts at zero so the bottom
            left corner of the image has coordinates (-0.5, -0.5).

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_phot_out_port = self.add_output_port(phot_out_tag)
        self.m_phot_in_port = None

        self.m_radius = radius
        self.m_position = position

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Computes the flux within a circular aperture for each
        frame and saves the values in the database.

        Returns
        -------
        NoneType
            None
        """

        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')
        self.m_radius /= pixscale

        if self.m_position is None:
            # Returns the center position as (y, x)
            self.m_position = center_subpixel(self.m_image_in_port[0, ])

            # Store the center position as (x, y)
            self.m_position = (self.m_position[1], self.m_position[0])

        print(f'Aperture position (x, y) = ({self.m_position[0]:.1f}, {self.m_position[1]:.1f})')
        print(f'Aperture radius (pixels) = ({self.m_radius:.1f})')

        # Position in CircularAperture is defined as (x, y)
        aperture = CircularAperture((self.m_position[0], self.m_position[1]), self.m_radius)

        self.apply_function_to_images(photometry,
                                      self.m_image_in_port,
                                      self.m_phot_out_port,
                                      'Aperture photometry',
                                      func_args=(aperture, ))

        self.m_phot_in_port = self.add_input_port(self.m_phot_out_port.tag)
        data = self.m_phot_in_port.get_all()

        print(f'Mean flux (counts) = {np.mean(data):.2f} +/- {np.std(data)/np.sqrt(data.size):.2f}')

        history = f'radius (pixels) = {self.m_radius:.3f}'
        self.m_phot_out_port.copy_attributes(self.m_image_in_port)
        self.m_phot_out_port.add_history('AperturePhotometryModule', history)
        self.m_phot_out_port.close_port()


class SystematicErrorModule(ProcessingModule):
    """
    Pipeline module for estimating the systematic error of the flux and position measurement.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 psf_in_tag: str,
                 offset_out_tag: str,
                 position: Tuple[float, float],
                 magnitude: float,
                 angles: Tuple[float, float, int] = (0., 359., 360),
                 psf_scaling: float = 1.,
                 merit: str = 'gaussian',
                 aperture: float = 0.1,
                 tolerance: float = 0.01,
                 pca_number: int = 10,
                 mask: Optional[Union[Tuple[float, float], Tuple[None, float],
                                      Tuple[float, None], Tuple[None, None]]] = None,
                 extra_rot: float = 0.,
                 residuals: str = 'median',
                 offset: Optional[float] = None) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry with the science images for which the systematic error is
            estimated.
        psf_in_tag : str
            Tag of the database entry with the PSF template that is used as fake planet. Can be
            either a single image or a stack of images equal in size to ``image_in_tag``.
        offset_out_tag : str
            Tag of the database entry at which the differences are stored between the injected and
            and retrieved values of the separation (arcsec), position angle (deg), contrast
            (mag), x position (pix), and y position (pix).
        position : tuple(float, float)
            Separation (arcsec) and position angle (deg) that are used to remove the planet signal.
            The separation is also used to estimate the systematic error.
        magnitude : float
            Magnitude that is used to remove the planet signal and estimate the systematic error.
        angles : tuple(float, float, int)
            The start, end, and number of the position angles (linearly sampled) that are used to
            estimate the systematic errors (default: 0., 359., 360). The endpoint is also included.
        psf_scaling : float
            Additional scaling factor of the planet flux (e.g., to correct for a neutral density
            filter). Should be a positive value.
        merit : str
            Figure of merit for the minimization ('hessian', 'gaussian', or 'poisson'). Either the
            determinant of the Hessian matrix is minimized ('hessian') or the flux of each pixel
            ('gaussian' or 'poisson'). For the latter case, the estimate noise is assumed to follow
            a Poisson (see Wertz et al. 2017) or Gaussian distribution (see Wertz et al. 2017 and
            Stolker et al. 2020).
        aperture : float
            Aperture radius (arcsec) that is used for measuring the figure of merit.
        tolerance : float
            Absolute error on the input parameters, position (pix) and contrast (mag), that is used
            as acceptance level for convergence. Note that only a single value can be specified
            which is used for both the position and flux so tolerance=0.1 will give a precision of
            0.1 mag and 0.1 pix. The tolerance on the output (i.e., the chi-square value) is set to
            np.inf so the condition is always met.
        pca_number : int
            Number of principal components (PCs) used for the PSF subtraction.
        mask : tuple(float, float), None
            Inner and outer mask radius (arcsec) which is applied before the PSF subtraction. Both
            elements of the tuple can be set to None.
        extra_rot : float
            Additional rotation angle of the images in clockwise direction (deg).
        residuals : str
            Method for combining the residuals ('mean', 'median', 'weighted', or 'clipped').
        offset : float, None
            Offset (pixels) by which the negative PSF may deviate from the positive injected PSF.
            No constraint on the position is applied if set to None. Only the contrast is optimized
            and the position is fixed to the injected value if ``offset=0``.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_tag = image_in_tag
        self.m_psf_in_tag = psf_in_tag

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_offset_out_port = self.add_output_port(offset_out_tag)

        self.m_position = position
        self.m_magnitude = magnitude
        self.m_angles = angles
        self.m_psf_scaling = psf_scaling
        self.m_merit = merit
        self.m_aperture = aperture
        self.m_tolerance = tolerance
        self.m_mask = mask
        self.m_extra_rot = extra_rot
        self.m_residuals = residuals
        self.m_pca_number = pca_number
        self.m_offset = offset

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Removes the planet signal, then artificial planets are injected
        (one at a time) at equally separated position angles and their position and contrast is
        determined with the :class:`~pynpoint.processing.fluxposition.SimplexMinimizationModule`.
        The differences between the injected and retrieved separation, position angle, and contrast
        are then stored as output.

        Returns
        -------
        NoneType
            None
        """

        print('Input parameters:')
        print(f'   - Number of principal components = {self.m_pca_number}')
        print(f'   - Figure of merit = {self.m_merit}')
        print(f'   - Residuals type = {self.m_residuals}')
        print(f'   - Absolute tolerance (pixels/mag) = {self.m_tolerance}')
        print(f'   - Maximum offset = {self.m_offset}')
        print(f'   - Aperture radius (arcsec) = {self.m_aperture}')

        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')
        image = self.m_image_in_port[0, ]

        module = FakePlanetModule(name_in=f'{self._m_name}_fake',
                                  image_in_tag=self.m_image_in_tag,
                                  psf_in_tag=self.m_psf_in_tag,
                                  image_out_tag=f'{self._m_name}_empty',
                                  position=(self.m_position[0],
                                            self.m_position[1]+self.m_extra_rot),
                                  magnitude=self.m_magnitude,
                                  psf_scaling=-self.m_psf_scaling)

        module.connect_database(self._m_data_base)
        module._m_output_ports[f'{self._m_name}_empty'].del_all_data()
        module._m_output_ports[f'{self._m_name}_empty'].del_all_attributes()
        module.run()

        sep = float(self.m_position[0])

        angles = np.linspace(self.m_angles[0], self.m_angles[1], self.m_angles[2], endpoint=True)

        print('Testing the following parameters:')
        print(f'   - Contrast (mag) = {self.m_magnitude:.2f}')
        print(f'   - Separation (mas) = {sep*1e3:.1f}')
        print(f'   - Position angle range (deg) = {angles[0]} - {angles[-1]}')

        if angles.size > 1:
            print(f'     in steps of {np.mean(np.diff(angles)):.2f} deg')

        # Image center (y, x) with subpixel accuracy
        im_center = center_subpixel(image)

        for i, ang in enumerate(angles):
            print(f'\nProcessing position angle: {ang} deg...')

            # Convert the polar coordiantes of the separation and position angle that is tested
            # into cartesian coordinates (y, x)
            planet_pos_yx = polar_to_cartesian(image, sep/pixscale, ang)
            planet_pos_xy = (planet_pos_yx[1], planet_pos_yx[0])

            # Convert the planet position to polar coordinates
            planet_sep_ang = cartesian_to_polar(im_center, planet_pos_yx[0], planet_pos_yx[1])

            # Change the separation units to arcsec
            planet_sep_ang = (planet_sep_ang[0]*pixscale, planet_sep_ang[1])

            # Inject the artifical planet

            module = FakePlanetModule(position=(planet_sep_ang[0],
                                                planet_sep_ang[1]+self.m_extra_rot),
                                      magnitude=self.m_magnitude,
                                      psf_scaling=self.m_psf_scaling,
                                      name_in=f'{self._m_name}_fake_{i}',
                                      image_in_tag=f'{self._m_name}_empty',
                                      psf_in_tag=self.m_psf_in_tag,
                                      image_out_tag=f'{self._m_name}_fake')

            module.connect_database(self._m_data_base)
            module._m_output_ports[f'{self._m_name}_fake'].del_all_data()
            module._m_output_ports[f'{self._m_name}_fake'].del_all_attributes()
            module.run()

            # Retrieve the position and contrast of the artificial planet

            module = SimplexMinimizationModule(position=planet_pos_xy,
                                               magnitude=self.m_magnitude,
                                               psf_scaling=-self.m_psf_scaling,
                                               name_in=f'{self._m_name}_fake_{i}',
                                               image_in_tag=f'{self._m_name}_fake',
                                               psf_in_tag=self.m_psf_in_tag,
                                               res_out_tag=f'{self._m_name}_simplex',
                                               flux_position_tag=f'{self._m_name}_fluxpos',
                                               merit=self.m_merit,
                                               aperture=self.m_aperture,
                                               sigma=0.,
                                               tolerance=self.m_tolerance,
                                               pca_number=self.m_pca_number,
                                               cent_size=self.m_mask[0],
                                               edge_size=self.m_mask[1],
                                               extra_rot=self.m_extra_rot,
                                               residuals=self.m_residuals,
                                               offset=self.m_offset)

            module.connect_database(self._m_data_base)
            module._m_output_ports[f'{self._m_name}_simplex'].del_all_data()
            module._m_output_ports[f'{self._m_name}_simplex'].del_all_attributes()
            module._m_output_ports[f'{self._m_name}_fluxpos'].del_all_data()
            module._m_output_ports[f'{self._m_name}_fluxpos'].del_all_attributes()
            module.run()

            # Add the input port to collect the results of SimplexMinimizationModule
            fluxpos_out_port = self.add_input_port(f'{self._m_name}_fluxpos')

            # Create a list with the offset between the injected and retrieved values of the
            # separation (arcsec), position angle (deg), contrast (mag), x position (pixels),
            # and y position (pixels).
            data = [planet_sep_ang[0] - fluxpos_out_port[-1, 2],  # Separation (arcsec)
                    planet_sep_ang[1] - fluxpos_out_port[-1, 3],  # Position angle (deg)
                    self.m_magnitude - fluxpos_out_port[-1, 4],  # Contrast (mag)
                    planet_pos_xy[0] - fluxpos_out_port[-1, 0],  # Position x (pixels)
                    planet_pos_xy[1] - fluxpos_out_port[-1, 1]]  # Position y (pixels)

            if data[1] > 180.:
                data[1] -= 360.

            elif data[1] < -180.:
                data[1] += 360.

            print(f'Offset: {data[0]*1e3:.2f} mas, {data[1]:.2f} deg, {data[2]:.2f} mag')

            self.m_offset_out_port.append(data, data_dim=2)

        offset_in_port = self.add_input_port(self.m_offset_out_port.tag)
        offsets = offset_in_port.get_all()

        sep_percen = np.percentile(offsets[:, 0], [16., 50., 84.])
        ang_percen = np.percentile(offsets[:, 1], [16., 50., 84.])
        mag_percen = np.percentile(offsets[:, 2], [16., 50., 84.])
        x_pos_percen = np.percentile(offsets[:, 3], [16., 50., 84.])
        y_pos_percen = np.percentile(offsets[:, 4], [16., 50., 84.])

        print('\nMedian offset and uncertainties:')

        print(f'   - Position x (pixels) = {x_pos_percen[1]:.2f} '
              f'(-{x_pos_percen[1]-x_pos_percen[0]:.2f} '
              f'+{x_pos_percen[2]-x_pos_percen[1]:.2f})')

        print(f'   - Position y (pixels) = {y_pos_percen[1]:.2f} '
              f'(-{y_pos_percen[1]-y_pos_percen[0]:.2f} '
              f'+{y_pos_percen[2]-y_pos_percen[1]:.2f})')

        print(f'   - Separation (mas) = {1e3*sep_percen[1]:.2f} '
              f'(-{1e3*sep_percen[1]-1e3*sep_percen[0]:.2f} '
              f'+{1e3*sep_percen[2]-1e3*sep_percen[1]:.2f})')

        print(f'   - Position angle (deg) = {ang_percen[1]:.2f} '
              f'(-{ang_percen[1]-ang_percen[0]:.2f} '
              f'+{ang_percen[2]-ang_percen[1]:.2f})')

        print(f'   - Contrast (mag) = {mag_percen[1]:.2f} '
              f'(-{mag_percen[1]-mag_percen[0]:.2f} '
              f'+{mag_percen[2]-mag_percen[1]:.2f})')

        history = f'sep = {self.m_position[0]:.3f}, ' \
                  f'pa = {self.m_position[1]:.1f}, ' \
                  f'mag = {self.m_magnitude:.1f}'

        self.m_offset_out_port.copy_attributes(self.m_image_in_port)
        self.m_offset_out_port.add_history('SystematicErrorModule', history)
        self.m_offset_out_port.close_port()



import os

import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.centering import StarAlignmentModule, ShiftImagesModule, \
                                          WaffleCenteringModule, FitCenterModule
from pynpoint.processing.extract import StarExtractionModule
from pynpoint.processing.resizing import AddLinesModule
from pynpoint.util.tests import create_config, create_star_data, create_waffle_data, \
                                remove_test_data


class TestCentering:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_star_data(self.test_dir+'star')
        create_waffle_data(self.test_dir+'waffle')
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(path=self.test_dir,
                         folders=['star', 'waffle'])

    def test_read_data(self) -> None:

        module = FitsReadingModule(name_in='read1',
                                   image_tag='star',
                                   input_dir=self.test_dir+'star',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)

        module = FitsReadingModule(name_in='read2',
                                   image_tag='waffle',
                                   input_dir=self.test_dir+'waffle',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)

        self.pipeline.run_module('read1')
        self.pipeline.run_module('read2')

        data = self.pipeline.get_data('star')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

        data = self.pipeline.get_data('waffle')
        assert np.sum(data) == pytest.approx(4.000000000000196, rel=self.limit, abs=0.)
        assert data.shape == (1, 101, 101)

    def test_star_extract(self) -> None:

        module = StarExtractionModule(name_in='extract1',
                                      image_in_tag='star',
                                      image_out_tag='extract1',
                                      index_out_tag='index',
                                      image_size=0.2,
                                      fwhm_star=0.1,
                                      position=None)

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('extract1')

        assert len(warning) == 3

        assert warning[0].message.args[0] == 'Can not store the attribute \'INSTRUMENT\' ' \
                                             'because the dataset \'index\' does not exist.'

        data = self.pipeline.get_data('extract1')
        assert np.sum(data) == pytest.approx(104.93318507061295, rel=self.limit, abs=0.)
        assert data.shape == (10, 9, 9)

    def test_star_align(self) -> None:

        module = StarAlignmentModule(name_in='align1',
                                     image_in_tag='extract1',
                                     ref_image_in_tag=None,
                                     image_out_tag='align1',
                                     accuracy=10,
                                     resize=2.,
                                     num_references=10,
                                     subframe=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('align1')

        data = self.pipeline.get_data('align1')
        assert np.sum(data) == pytest.approx(104.7074742320535, rel=self.limit, abs=0.)
        assert data.shape == (10, 18, 18)

    def test_star_align_subframe(self) -> None:

        module = StarAlignmentModule(name_in='align2',
                                     image_in_tag='extract1',
                                     ref_image_in_tag=None,
                                     image_out_tag='align2',
                                     accuracy=10,
                                     resize=None,
                                     num_references=10,
                                     subframe=0.1)

        self.pipeline.add_module(module)
        self.pipeline.run_module('align2')

        data = self.pipeline.get_data('align2')
        assert np.sum(data) == pytest.approx(104.39031104541652, rel=self.limit, abs=0.)
        assert data.shape == (10, 9, 9)

    def test_star_align_ref(self) -> None:

        module = StarAlignmentModule(name_in='align3',
                                     image_in_tag='extract1',
                                     ref_image_in_tag='align2',
                                     image_out_tag='align3',
                                     accuracy=10,
                                     resize=None,
                                     num_references=10,
                                     subframe=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('align3')

        data = self.pipeline.get_data('align3')
        assert np.sum(data) == pytest.approx(104.46997194330757, rel=self.limit, abs=0.)
        assert data.shape == (10, 9, 9)

    def test_star_align_number_ref(self) -> None:

        module = StarAlignmentModule(name_in='align4',
                                     image_in_tag='extract1',
                                     ref_image_in_tag='align2',
                                     image_out_tag='align4',
                                     accuracy=10,
                                     resize=None,
                                     num_references=20,
                                     subframe=None)

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('align4')

        assert len(warning) == 1

        assert warning[0].message.args[0] == 'Number of available images (10) is smaller than ' \
                                             'num_references (20). Using all available images ' \
                                             'instead.'

        data = self.pipeline.get_data('align4')
        assert np.sum(data) == pytest.approx(104.46997194330757, rel=self.limit, abs=0.)
        assert data.shape == (10, 9, 9)

    def test_shift_images_spline(self) -> None:

        module = ShiftImagesModule(shift_xy=(1., 2.),
                                   interpolation='spline',
                                   name_in='shift1',
                                   image_in_tag='align1',
                                   image_out_tag='shift')

        self.pipeline.add_module(module)
        self.pipeline.run_module('shift1')

        data = self.pipeline.get_data('shift')
        assert np.sum(data) == pytest.approx(104.20425101355244, rel=self.limit, abs=0.)
        assert data.shape == (10, 18, 18)

    def test_shift_images_fft(self) -> None:

        module = ShiftImagesModule(shift_xy=(1., 2.),
                                   interpolation='fft',
                                   name_in='shift2',
                                   image_in_tag='align1',
                                   image_out_tag='shift_fft')

        self.pipeline.add_module(module)
        self.pipeline.run_module('shift2')

        data = self.pipeline.get_data('shift_fft')
        assert np.sum(data) == pytest.approx(104.7074742320535, rel=self.limit, abs=0.)
        assert data.shape == (10, 18, 18)

    def test_waffle_center_odd(self) -> None:

        module = AddLinesModule(name_in='add',
                                image_in_tag='star',
                                image_out_tag='star_add',
                                lines=(45, 45, 45, 45))

        self.pipeline.add_module(module)
        self.pipeline.run_module('add')

        data = self.pipeline.get_data('star_add')
        assert np.sum(data) == pytest.approx(105.54278879805278, rel=self.limit, abs=0.)
        assert data.shape == (10, 101, 101)

        module = WaffleCenteringModule(size=0.2,
                                       center=(50, 50),
                                       name_in='waffle',
                                       image_in_tag='star_add',
                                       center_in_tag='waffle',
                                       image_out_tag='center',
                                       radius=35.,
                                       pattern='x',
                                       sigma=0.05)

        self.pipeline.add_module(module)

        with pytest.warns(DeprecationWarning) as warning:
            self.pipeline.run_module('waffle')

        assert len(warning) == 1

        assert warning[0].message.args[0] == 'The \'pattern\' parameter will be deprecated in a ' \
                                             'future release. Please Use the \'angle\' ' \
                                             'parameter instead and set it to 45.0 degrees.'

        data = self.pipeline.get_data('center')
        assert np.sum(data) == pytest.approx(104.93318507061295, rel=self.limit, abs=0.)
        assert data.shape == (10, 9, 9)

        attr = self.pipeline.get_attribute('center', 'History: WaffleCenteringModule')
        assert attr == '[x, y] = [50.0, 50.0]'

    def test_waffle_center_even(self) -> None:

        module = AddLinesModule(name_in='add1',
                                image_in_tag='star_add',
                                image_out_tag='star_even',
                                lines=(0, 1, 0, 1))

        self.pipeline.add_module(module)
        self.pipeline.run_module('add1')

        data = self.pipeline.get_data('star_even')
        assert np.sum(data) == pytest.approx(105.54278879805275, rel=self.limit, abs=0.)
        assert data.shape == (10, 102, 102)

        module = AddLinesModule(name_in='add2',
                                image_in_tag='waffle',
                                image_out_tag='waffle_even',
                                lines=(0, 1, 0, 1))

        self.pipeline.add_module(module)
        self.pipeline.run_module('add2')

        data = self.pipeline.get_data('waffle_even')
        assert np.sum(data) == pytest.approx(4.000000000000195, rel=self.limit, abs=0.)
        assert data.shape == (1, 102, 102)

        module = ShiftImagesModule(shift_xy=(0.5, 0.5),
                                   interpolation='spline',
                                   name_in='shift3',
                                   image_in_tag='star_even',
                                   image_out_tag='star_shift')

        self.pipeline.add_module(module)
        self.pipeline.run_module('shift3')

        data = self.pipeline.get_data('star_shift')
        assert np.sum(data) == pytest.approx(105.54278879805274, rel=self.limit, abs=0.)
        assert data.shape == (10, 102, 102)

        module = ShiftImagesModule(shift_xy=(0.5, 0.5),
                                   interpolation='spline',
                                   name_in='shift4',
                                   image_in_tag='waffle_even',
                                   image_out_tag='waffle_shift')

        self.pipeline.add_module(module)
        self.pipeline.run_module('shift4')

        data = self.pipeline.get_data('waffle_shift')
        assert np.sum(data) == pytest.approx(4.000000000000194, rel=self.limit, abs=0.)
        assert data.shape == (1, 102, 102)

        module = WaffleCenteringModule(size=0.2,
                                       center=(50, 50),
                                       name_in='waffle_even',
                                       image_in_tag='star_shift',
                                       center_in_tag='waffle_shift',
                                       image_out_tag='center_even',
                                       radius=35.,
                                       pattern='x',
                                       sigma=0.05)

        self.pipeline.add_module(module)

        with pytest.warns(DeprecationWarning) as warning:
            self.pipeline.run_module('waffle_even')

        assert len(warning) == 1

        assert warning[0].message.args[0] == 'The \'pattern\' parameter will be deprecated in a ' \
                                             'future release. Please Use the \'angle\' ' \
                                             'parameter instead and set it to 45.0 degrees.'

        data = self.pipeline.get_data('center_even')
        assert np.sum(data) == pytest.approx(105.22695036281449, rel=self.limit, abs=0.)
        assert data.shape == (10, 9, 9)

        attr = self.pipeline.get_attribute('center_even', 'History: WaffleCenteringModule')
        assert attr == '[x, y] = [50.5, 50.5]'

    def test_fit_center_full(self) -> None:

        module = FitCenterModule(name_in='fit1',
                                 image_in_tag='shift',
                                 fit_out_tag='fit_full',
                                 mask_out_tag='mask',
                                 method='full',
                                 mask_radii=(None, 0.1),
                                 sign='positive',
                                 model='gaussian',
                                 guess=(1., 2., 3., 3., 0.01, 0., 0.))

        self.pipeline.add_module(module)
        self.pipeline.run_module('fit1')

        data = self.pipeline.get_data('fit_full')

        assert np.mean(data[:, 0]) == pytest.approx(0.94, rel=0., abs=0.05)
        assert np.mean(data[:, 2]) == pytest.approx(2.07, rel=0., abs=0.05)
        assert np.mean(data[:, 4]) == pytest.approx(0.0676, rel=0., abs=0.05)
        assert np.mean(data[:, 6]) == pytest.approx(0.0665, rel=0., abs=0.05)
        assert np.mean(data[:, 8]) == pytest.approx(0.24, rel=0., abs=0.05)
        assert data.shape == (10, 14)

        data = self.pipeline.get_data('mask')
        assert np.sum(data) == pytest.approx(103.45599730750453, rel=self.limit, abs=0.)
        assert data.shape == (10, 18, 18)

    def test_fit_center_mean(self) -> None:

        module = FitCenterModule(name_in='fit2',
                                 image_in_tag='shift',
                                 fit_out_tag='fit_mean',
                                 mask_out_tag=None,
                                 method='mean',
                                 mask_radii=(None, 0.1),
                                 sign='positive',
                                 model='moffat',
                                 guess=(1., 2., 3., 3., 0.01, 0., 0., 1.))

        self.pipeline.add_module(module)
        self.pipeline.run_module('fit2')

        data = self.pipeline.get_data('fit_mean')
        assert np.mean(data[:, 0]) == pytest.approx(0.94, rel=0., abs=0.01)
        assert np.mean(data[:, 2]) == pytest.approx(2.059, rel=0., abs=0.01)
        assert np.mean(data[:, 4]) == pytest.approx(0.083, rel=0., abs=0.01)
        assert np.mean(data[:, 6]) == pytest.approx(0.08, rel=0., abs=0.01)
        assert np.mean(data[:, 8]) == pytest.approx(0.242, rel=0., abs=0.01)
        assert data.shape == (10, 16)

    def test_shift_images_tag(self) -> None:

        module = ShiftImagesModule(shift_xy='fit_full',
                                   interpolation='spline',
                                   name_in='shift5',
                                   image_in_tag='shift',
                                   image_out_tag='shift_tag_1')

        self.pipeline.add_module(module)
        self.pipeline.run_module('shift5')

        data = self.pipeline.get_data('shift_tag_1')
        assert np.sum(data) == pytest.approx(103.76410960085425, rel=1e-6, abs=0.9)
        assert data.shape == (10, 18, 18)

    def test_shift_images_tag_mean(self) -> None:

        module = ShiftImagesModule(shift_xy='fit_mean',
                                   interpolation='spline',
                                   name_in='shift6',
                                   image_in_tag='shift',
                                   image_out_tag='shift_tag_2')

        self.pipeline.add_module(module)
        self.pipeline.run_module('shift6')

        data = self.pipeline.get_data('shift_tag_2')
        assert np.sum(data) == pytest.approx(103.42285439876926, rel=1e-6, abs=0.)
        assert data.shape == (10, 18, 18)



"""
Modules for reading attributes from a FITS or ASCII file.
"""

import os
import warnings

from typing import Optional

import numpy as np

from astropy.io import fits
from typeguard import typechecked

from pynpoint.core.attributes import get_attributes
from pynpoint.core.processing import ReadingModule


class AttributeReadingModule(ReadingModule):
    """
    Module for reading a list of values from a FITS or ASCII file and appending them as a non-static
    attributes to a dataset.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 data_tag: str,
                 file_name: str,
                 attribute: str,
                 input_dir: Optional[str] = None,
                 overwrite: bool = False) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        data_tag : str
            Tag of the database entry to which the attribute is written.
        file_name : str
            Name of the input file with the attribute value. Should be equal in size to the number
            of images in ``data_tag``. In case the ``file_name`` is ending with ``.fits``, then a
            FITS file is read. Otherwise, a single column of values is expected in an ASCII file.
        file_name : str
            Name of the input file with a list of values.
        attribute : str
            Name of the attribute as to be written in the database.
        input_dir : str, None
            Input directory where the input file is located. If not specified the Pypeline default
            directory is used.
        overwrite : bool
            Overwrite if the attribute is already exists.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in, input_dir=input_dir)

        self.m_data_port = self.add_output_port(data_tag)

        self.m_file_name = file_name
        self.m_attribute = attribute
        self.m_overwrite = overwrite

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Reads a list of values from a FITS or ASCII file and writes them
        as non-static attribute to a dataset.

        Returns
        -------
        NoneType
            None
        """

        print('Reading attribute data...', end='')

        attributes = get_attributes()

        if self.m_attribute not in attributes:
            raise ValueError(f'\'{self.m_attribute}\' is not a valid attribute.')

        if self.m_file_name.endswith('fits'):
            values = fits.getdata(os.path.join(self.m_input_location, self.m_file_name))
        else:
            values = np.loadtxt(os.path.join(self.m_input_location, self.m_file_name),
                                dtype=attributes[self.m_attribute]['type'])

        if values.ndim != 1:
            raise ValueError(f'The input file {self.m_file_name} should contain a 1D list with '
                             f'attributes.')

        status = self.m_data_port.check_non_static_attribute(self.m_attribute, values)

        if status == 1:
            self.m_data_port.add_attribute(self.m_attribute, values, static=False)

        elif status == -1 and self.m_overwrite:
            self.m_data_port.add_attribute(self.m_attribute, values, static=False)

        elif status == -1 and not self.m_overwrite:
            warnings.warn(f'The attribute \'{self.m_attribute}\' is already present. Set the '
                          f'\'overwrite\' parameter to True in order to overwrite the values with '
                          f'{self.m_file_name}.')

        elif status == 0:
            warnings.warn(f'The \'{self.m_attribute}\' attribute is already present and '
                          f'contains the same values as are present in {self.m_file_name}.')

        print(' [DONE]')

        self.m_data_port.close_port()


class ParangReadingModule(ReadingModule):
    """
    Module for reading a list of parallactic angles from a FITS or ASCII file.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 data_tag: str,
                 file_name: str,
                 input_dir: Optional[str] = None,
                 overwrite: bool = False) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        data_tag : str
            Tag of the database entry to which the ``PARANG`` attribute is written.
        file_name : str
            Name of the input file with the parallactic angles (deg). Should be equal in size
            to the number of images in ``data_tag``. In case the ``file_name`` is ending with
            ``.fits``, then a FITS file is read. Otherwise, a single column of values is expected
            in an ASCII file.
        input_dir : str, None
            Input directory where the input file is located. If not specified the Pypeline default
            directory is used.
        overwrite : bool
            Overwrite if the ``PARANG`` attribute already exists.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in, input_dir=input_dir)

        self.m_data_port = self.add_output_port(data_tag)

        self.m_file_name = file_name
        self.m_overwrite = overwrite

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Reads the parallactic angles from a FITS or ASCII file and
        writes the values as non-static attribute (``PARANG``) to the database tag.

        Returns
        -------
        NoneType
            None
        """

        print('Reading parallactic angles...', end='')

        if self.m_file_name.endswith('fits'):
            parang = fits.getdata(os.path.join(self.m_input_location, self.m_file_name))
        else:
            parang = np.loadtxt(os.path.join(self.m_input_location, self.m_file_name))

        print(' [DONE]')

        if parang.ndim != 1:
            raise ValueError(f'The input file {self.m_file_name} should contain a 1D data set with '
                             f'the parallactic angles.')

        print(f'Number of angles: {parang.size}')
        print(f'Rotation range: {parang[0]:.2f} -> {parang[-1]:.2f} deg')

        status = self.m_data_port.check_non_static_attribute('PARANG', parang)

        if status == 1:
            self.m_data_port.add_attribute('PARANG', parang, static=False)

        elif status == -1 and self.m_overwrite:
            self.m_data_port.add_attribute('PARANG', parang, static=False)

        elif status == -1 and not self.m_overwrite:
            warnings.warn(f'The PARANG attribute is already present. Set the \'overwrite\' '
                          f'parameter to True in order to overwrite the values with '
                          f'{self.m_file_name}.')

        elif status == 0:
            warnings.warn(f'The PARANG attribute is already present and contains the same values '
                          f'as are present in {self.m_file_name}.')

        self.m_data_port.close_port()


class WavelengthReadingModule(ReadingModule):
    """
    Module for reading a list of wavelengths from a FITS or ASCII file.
    """

    __author__ = 'Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 data_tag: str,
                 file_name: str,
                 input_dir: Optional[str] = None,
                 overwrite: bool = False) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        data_tag : str
            Tag of the database entry to which the ``WAVELENGTH`` attribute is written.
        file_name : str
            Name of the input file with the wavelengths (a.u.). Should be equal in size
            to the number of images in ``data_tag``. In case the ``file_name`` is ending with
            ``.fits``, then a FITS file is read. Otherwise, a single column of values is expected
            in an ASCII file.
        input_dir : str, None
            Input directory where the input file is located. If not specified the Pypeline default
            directory is used.
        overwrite : bool
            Overwrite if the ``WAVELENGTH`` attribute already exists.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in, input_dir=input_dir)

        self.m_data_port = self.add_output_port(data_tag)

        self.m_file_name = file_name
        self.m_overwrite = overwrite

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Reads the parallactic angles from a FITS or ASCII file and writes
        the values as non-static attribute (``WAVELENGTH``) to the database tag.

        Returns
        -------
        NoneType
            None
        """

        print('Reading wavelengths...', end='')

        if self.m_file_name.endswith('fits'):
            wavelength = fits.getdata(os.path.join(self.m_input_location, self.m_file_name))
        else:
            wavelength = np.loadtxt(os.path.join(self.m_input_location, self.m_file_name))

        print(' [DONE]')

        if wavelength.ndim != 1:
            raise ValueError(f'The input file {self.m_file_name} should contain a 1D data set with '
                             f'the wavelengths.')

        print(f'Number of wavelengths: {wavelength.size}')
        print(f'Wavelength range: {wavelength[0]:.2f} - {wavelength[-1]:.2f}')

        status = self.m_data_port.check_non_static_attribute('WAVELENGTH', wavelength)

        if status == 1:
            self.m_data_port.add_attribute('WAVELENGTH', wavelength, static=False)

        elif status == -1 and self.m_overwrite:
            self.m_data_port.add_attribute('WAVELENGTH', wavelength, static=False)

        elif status == -1 and not self.m_overwrite:
            warnings.warn(f'The WAVELENGTH attribute is already present. Set the \'overwrite\' '
                          f'parameter to True in order to overwrite the values with '
                          f'{self.m_file_name}.')

        elif status == 0:
            warnings.warn(f'The WAVELENGTH attribute is already present and contains the same '
                          f'values as are present in {self.m_file_name}.')

        self.m_data_port.close_port()



import os

import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.psfpreparation import PSFpreparationModule, AngleInterpolationModule, \
                                               AngleCalculationModule, SDIpreparationModule, \
                                               SortParangModule
from pynpoint.util.tests import create_config, create_star_data, create_ifs_data, remove_test_data


class TestPsfPreparation:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_star_data(self.test_dir+'prep')
        create_ifs_data(self.test_dir+'prep_ifs')
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=['prep', 'prep_ifs'])

    def test_read_data(self) -> None:

        module = FitsReadingModule(name_in='read',
                                   image_tag='read',
                                   input_dir=self.test_dir+'prep')

        self.pipeline.add_module(module)
        self.pipeline.run_module('read')

        data = self.pipeline.get_data('read')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

        module = FitsReadingModule(name_in='read_ifs',
                                   image_tag='read_ifs',
                                   input_dir=self.test_dir+'prep_ifs',
                                   ifs_data=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('read_ifs')

        data = self.pipeline.get_data('read_ifs')
        assert np.sum(data) == pytest.approx(749.8396528807369, rel=self.limit, abs=0.)
        assert data.shape == (3, 10, 21, 21)

    def test_angle_interpolation(self) -> None:

        module = AngleInterpolationModule(name_in='angle1',
                                          data_tag='read')

        self.pipeline.add_module(module)
        self.pipeline.run_module('angle1')

        data = self.pipeline.get_data('header_read/PARANG')
        assert np.sum(data) == pytest.approx(900., rel=self.limit, abs=0.)
        assert data.shape == (10, )

    def test_angle_calculation(self) -> None:

        self.pipeline.set_attribute('read', 'LATITUDE', -25.)
        self.pipeline.set_attribute('read', 'LONGITUDE', -70.)
        self.pipeline.set_attribute('read', 'DIT', 1.)

        self.pipeline.set_attribute('read', 'RA', (90., 90., 90., 90.), static=False)
        self.pipeline.set_attribute('read', 'DEC', (-51., -51., -51., -51.), static=False)
        self.pipeline.set_attribute('read', 'PUPIL', (90., 90., 90., 90.), static=False)

        date = ('2012-12-01T07:09:00.0000', '2012-12-01T07:09:01.0000',
                '2012-12-01T07:09:02.0000', '2012-12-01T07:09:03.0000')

        self.pipeline.set_attribute('read', 'DATE', date, static=False)

        module = AngleCalculationModule(instrument='NACO',
                                        name_in='angle2',
                                        data_tag='read')

        self.pipeline.add_module(module)
        self.pipeline.run_module('angle2')

        data = self.pipeline.get_data('header_read/PARANG')
        assert np.sum(data) == pytest.approx(-550.2338293743467, rel=self.limit, abs=0.)
        assert data.shape == (10, )

        self.pipeline.set_attribute('read', 'RA', (60000.0, 60000.0, 60000.0, 60000.0),
                                    static=False)

        self.pipeline.set_attribute('read', 'DEC', (-510000., -510000., -510000., -510000.),
                                    static=False)

        module = AngleCalculationModule(instrument='SPHERE/IRDIS',
                                        name_in='angle3',
                                        data_tag='read')

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('angle3')

        warning_0 = 'For SPHERE data it is recommended to use the header keyword \'ESO INS4 ' \
                    'DROT2 RA\' to specify the object\'s right ascension. The input will be ' \
                    'parsed accordingly. Using the regular \'RA\' keyword will lead to wrong ' \
                    'parallactic angles.'

        warning_1 = 'For SPHERE data it is recommended to use the header keyword \'ESO INS4 ' \
                    'DROT2 DEC\' to specify the object\'s declination. The input will be parsed ' \
                    'accordingly. Using the regular \'DEC\' keyword will lead to wrong ' \
                    'parallactic angles.'

        if len(warning) == 2:
            assert warning[0].message.args[0] == warning_0
            assert warning[1].message.args[0] == warning_1

        data = self.pipeline.get_data('header_read/PARANG')
        assert np.sum(data) == pytest.approx(1704.2202367435675, rel=self.limit, abs=0.)
        assert data.shape == (10, )

        module = AngleCalculationModule(instrument='SPHERE/IFS',
                                        name_in='angle4',
                                        data_tag='read')

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('angle4')

        warning_0 = 'AngleCalculationModule has not been tested for SPHERE/IFS data.'

        warning_1 = 'For SPHERE data it is recommended to use the header keyword \'ESO INS4 ' \
                    'DROT2 RA\' to specify the object\'s right ascension. The input will be ' \
                    'parsed accordingly. Using the regular \'RA\' keyword will lead to wrong ' \
                    'parallactic angles.'

        warning_2 = 'For SPHERE data it is recommended to use the header keyword \'ESO INS4 ' \
                    'DROT2 DEC\' to specify the object\'s declination. The input will be parsed ' \
                    'accordingly. Using the regular \'DEC\' keyword will lead to wrong ' \
                    'parallactic angles.'

        if len(warning) == 3:
            assert warning[0].message.args[0] == warning_0
            assert warning[1].message.args[0] == warning_1
            assert warning[2].message.args[0] == warning_2

        data = self.pipeline.get_data('header_read/PARANG')
        assert np.sum(data) == pytest.approx(-890.8506514377593, rel=self.limit, abs=0.)
        assert data.shape == (10, )

    def test_angle_sort(self) -> None:

        index = self.pipeline.get_data('header_read/INDEX')
        self.pipeline.set_attribute('read', 'INDEX', index[::-1], static=False)

        module = SortParangModule(name_in='sort1',
                                  image_in_tag='read',
                                  image_out_tag='read_sorted')

        self.pipeline.add_module(module)
        self.pipeline.run_module('sort1')
        self.pipeline.set_attribute('read', 'INDEX', index, static=False)

        parang = self.pipeline.get_data('header_read/PARANG')[::-1]
        parang_sort = self.pipeline.get_data('header_read_sorted/PARANG')
        assert np.sum(parang) == pytest.approx(np.sum(parang_sort), rel=self.limit, abs=0.)

        parang_set = [0., 1., 2., 3., 4., 5., 6., 7., 8., 9.]
        self.pipeline.set_attribute('read_ifs', 'PARANG', parang_set, static=False)

        data = self.pipeline.get_data('read_sorted')
        assert np.sum(data[0]) == pytest.approx(9.71156815235485, rel=self.limit, abs=0.)

    def test_angle_sort_ifs(self) -> None:

        index = self.pipeline.get_data('header_read_ifs/INDEX')
        self.pipeline.set_attribute('read_ifs', 'INDEX', index[::-1], static=False)

        module = SortParangModule(name_in='sort2',
                                  image_in_tag='read_ifs',
                                  image_out_tag='read_ifs_sorted')

        self.pipeline.add_module(module)
        self.pipeline.run_module('sort2')
        self.pipeline.set_attribute('read_ifs', 'INDEX', index, static=False)

        parang = self.pipeline.get_data('header_read_ifs/PARANG')[::-1]
        parang_sort = self.pipeline.get_data('header_read_ifs_sorted/PARANG')
        assert np.sum(parang) == pytest.approx(np.sum(parang_sort), rel=self.limit, abs=0.)

        data = self.pipeline.get_data('read_ifs_sorted')
        assert np.sum(data[0, 0]) == pytest.approx(21.185139976163477, rel=self.limit, abs=0.)

    def test_angle_interpolation_mismatch(self) -> None:

        self.pipeline.set_attribute('read', 'NDIT', [9, 9, 9, 9], static=False)

        module = AngleInterpolationModule(name_in='angle5',
                                          data_tag='read')

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('angle5')

        warning_0 = 'There is a mismatch between the NDIT and NFRAMES values. The parallactic ' \
                    'angles are calculated with a linear interpolation by using NFRAMES steps. ' \
                    'A frame selection should be applied after the parallactic angles are ' \
                    'calculated.'

        if len(warning) == 1:
            assert warning[0].message.args[0] == warning_0

        data = self.pipeline.get_data('header_read/PARANG')
        assert np.sum(data) == pytest.approx(900., rel=self.limit, abs=0.)
        assert data.shape == (10, )

    def test_psf_preparation_norm_mask(self) -> None:

        module = PSFpreparationModule(name_in='prep1',
                                      image_in_tag='read',
                                      image_out_tag='prep1',
                                      mask_out_tag='mask1',
                                      norm=True,
                                      cent_size=0.1,
                                      edge_size=1.0)

        self.pipeline.add_module(module)
        self.pipeline.run_module('prep1')

        data = self.pipeline.get_data('prep1')
        assert np.sum(data) == pytest.approx(-1.5844830188044685, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

        data = self.pipeline.get_data('mask1')
        assert np.sum(data) == pytest.approx(52, rel=self.limit, abs=0.)
        assert data.shape == (11, 11)

    def test_psf_preparation_none(self) -> None:

        module = PSFpreparationModule(name_in='prep2',
                                      image_in_tag='read',
                                      image_out_tag='prep2',
                                      mask_out_tag='mask2',
                                      norm=False,
                                      cent_size=None,
                                      edge_size=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('prep2')

        data = self.pipeline.get_data('prep2')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_psf_preparation_no_mask_out(self) -> None:

        module = PSFpreparationModule(name_in='prep3',
                                      image_in_tag='read',
                                      image_out_tag='prep3',
                                      mask_out_tag=None,
                                      norm=False,
                                      cent_size=None,
                                      edge_size=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('prep3')

        data = self.pipeline.get_data('prep3')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_psf_preparation_sdi(self) -> None:

        module = PSFpreparationModule(name_in='prep4',
                                      image_in_tag='read_ifs',
                                      image_out_tag='prep4',
                                      mask_out_tag=None,
                                      norm=False,
                                      cent_size=None,
                                      edge_size=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('prep4')

        data = self.pipeline.get_data('prep4')
        assert np.sum(data) == pytest.approx(749.8396528807369, rel=self.limit, abs=0.)
        assert data.shape == (3, 10, 21, 21)

    def test_sdi_preparation(self) -> None:

        module = SDIpreparationModule(name_in='sdi',
                                      wavelength=(0.65, 0.6),
                                      width=(0.1, 0.5),
                                      image_in_tag='read',
                                      image_out_tag='sdi')

        self.pipeline.add_module(module)
        self.pipeline.run_module('sdi')

        data = self.pipeline.get_data('sdi')
        assert np.sum(data) == pytest.approx(21.084666133914183, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

        attribute = self.pipeline.get_attribute('sdi', 'History: SDIpreparationModule')
        assert attribute == '(line, continuum) = (0.65, 0.6)'



"""
Functions for Pypeline modules.
"""

import sys
import time
import math
import cmath
import warnings

from typing import Dict, Optional, Tuple, TYPE_CHECKING, Union

import numpy as np

from typeguard import typechecked

# The following is needed to avoid circular dependencies:
# The PypelineModule uses methods from module.py, but  methods in this file also use PypelineModule
# for their type hints. If we were to simply import the PypelineModule class here, this circular
# dependency would lead to an ImportError at runtime. By using the TYPE_CHECKING flag, we can avoid
# this: because TYPE_CHECKING is always False at runtime, there are no ImportErrors, while the
# PypelineModule is still available for static type checkers (i.e., not for the typeguard library).
# In Python 3.7, this problem can be circumvented more elegantly by using:
#   >>> from __future__ import annotations
# This changes the behavior of type hints such that they are no longer evaluated at definition time,
# see also PEP 563. In Python 4.0, this is supposed to become the new default behavior.
if TYPE_CHECKING:
    from pynpoint.core.processing import PypelineModule


@typechecked
def progress(current: int,
             total: int,
             message: str,
             start_time: Optional[float] = None) -> None:
    """
    Function to show and update the progress as standard output.

    Parameters
    ----------
    current : int
        Current index.
    total : int
        Total index number.
    message : str
        Message that is printed.
    start_time : float, None, optional
        Start time in seconds. Not used if set to None.

    Returns
    -------
    NoneType
        None
    """

    @typechecked
    def time_string(delta_time: float) -> str:
        """
        Converts to input time in seconds to a string which displays as hh:mm:ss.

        Parameters
        ----------
        delta_time : float
            Input time in seconds.

        Returns
        -------
        str:
            String with the formatted time.
        """

        hours = int(delta_time / 3600.)
        minutes = int((delta_time % 3600.) / 60.)
        seconds = int(delta_time % 60.)

        return f'{hours:>02}:{minutes:>02}:{seconds:>02}'

    fraction = float(current) / float(total)
    percentage = 100.*fraction

    if start_time is None:
        sys.stdout.write(f'\r{message} {percentage:4.1f}% \r')

    else:
        if fraction > 0. and current+1 != total:
            time_taken = time.time() - start_time
            time_left = time_taken / fraction * (1. - fraction)
            sys.stdout.write(f'{message} {percentage:4.1f}% - ETA: {time_string(time_left)}\r')

    if current+1 == total:
        sys.stdout.write((29 + len(message)) * ' ' + '\r')
        sys.stdout.write(message+' [DONE]\n')

    sys.stdout.flush()


@typechecked
def memory_frames(memory: Union[int, np.int64],
                  nimages: int) -> np.ndarray:
    """
    Function to subdivide the input images is in quantities of MEMORY.

    Parameters
    ----------
    memory : int
        Number of images that is simultaneously loaded into the memory.
    nimages : int
        Number of images in the stack.

    Returns
    -------
    numpy.ndarray
        Array with the indices where a stack of images is subdivided.
    """

    if memory == 0 or memory >= nimages:
        frames = np.asarray([0, nimages])

    else:
        frames = np.linspace(start=0,
                             stop=nimages - nimages % memory,
                             num=int(float(nimages)/float(memory))+1,
                             endpoint=True,
                             dtype=np.int64)

        if nimages % memory > 0:
            frames = np.append(frames, nimages)

    return frames


@typechecked
def angle_average(angles: np.ndarray) -> float:
    """
    Function to calculate the average value of a list of angles.

    Parameters
    ----------
    angles : numpy.ndarray
        Parallactic angles (deg).

    Returns
    -------
    float
        Average angle (deg).
    """

    cmath_rect = sum(cmath.rect(1, math.radians(ang)) for ang in angles)
    cmath_phase = cmath.phase(cmath_rect/len(angles))

    return math.degrees(cmath_phase)


@typechecked
def angle_difference(angle_1: float,
                     angle_2: float) -> float:
    """
    Function to calculate the difference between two  angles.

    Parameters
    ----------
    angle_1 : float
        First angle (deg).
    angle_2 : float
        Second angle (deg).

    Returns
    -------
    float
        Angle difference (deg).
    """

    angle_diff = (angle_1-angle_2) % 360.

    if angle_diff >= 180.:
        angle_diff -= 360.

    return angle_diff


@typechecked
def stack_angles(memory: Union[int, np.int64],
                 parang: np.ndarray,
                 max_rotation: float) -> np.ndarray:
    """
    Function to subdivide the input images is in quantities of MEMORY with a restriction on the
    maximum field rotation across a subset of images.

    Parameters
    ----------
    memory : int
        Number of images that is simultaneously loaded into the memory.
    parang : numpy.ndarray
        Parallactic angles (deg).
    max_rotation : float
        Maximum field rotation (deg).

    Returns
    -------
    numpy.ndarray
        Array with the indices where a stack of images is subdivided.
    """

    warnings.warn('Testing of util.module.stack_angles has been limited, please use carefully.')

    nimages = parang.size

    if memory == 0 or memory >= nimages:
        frames = [0, nimages]

    else:
        frames = [0, ]
        parang_start = parang[0]
        im_count = 0

        for i in range(1, parang.size):
            abs_start_diff = abs(angle_difference(parang_start, parang[i-1]))
            abs_current_diff = abs(angle_difference(parang[i], parang[i-1]))

            if abs_start_diff > max_rotation or abs_current_diff > max_rotation:
                frames.append(i)
                parang_start = parang[i]
                im_count = 0

            else:
                im_count += 1

                if im_count == memory:
                    frames.append(i)

                    if i < parang.size-1:
                        parang_start = parang[i+1]
                        im_count = 0

        if frames[-1] != nimages:
            frames.append(nimages)

    return np.asarray(frames)


@typechecked
def update_arguments(index: int,
                     nimages: int,
                     args_in: Optional[tuple]) -> Optional[tuple]:
    """
    Function to update the arguments of an input function. Specifically, arguments which contain an
    array with the first dimension equal in size to the total number of images will be substituted
    by the array element of the image index.

    Parameters
    ----------
    index : int
        Image index in the stack.
    nimages : int
        Total number of images in the stack.
    args_in : tuple, None
        Function arguments that have to be updated.

    Returns
    -------
    tuple, None
        Updated function arguments.
    """

    if args_in is None:
        args_out = None

    else:
        args_out = []

        for item in args_in:
            if isinstance(item, np.ndarray) and item.shape[0] == nimages:
                args_out.append(item[index])

            else:
                args_out.append(item)

        args_out = tuple(args_out)

    return args_out


# This function *cannot* be decorated with @typechecked, because the typeguard library checks type
# hints at *runtime*, when PypelineModule is not available without causing circular dependencies.
def module_info(pipeline_module: 'PypelineModule') -> None:
    """
    Function to print the module name.

    Parameters
    ----------
    pipeline_module : PypelineModule
        Pipeline module.

    Returns
    -------
    NoneType
        None
    """

    module_name = type(pipeline_module).__name__
    str_length = len(module_name)

    print('\n' + str_length * '-')
    print(module_name)
    print(str_length * '-' + '\n')
    print(f'Module name: {pipeline_module._m_name}')


# This function *cannot* be decorated with @typechecked, because the typeguard library checks type
# hints at *runtime*, when PypelineModule is not available without causing circular dependencies.
def input_info(pipeline_module: 'PypelineModule') -> None:
    """
    Function to print information about the input data.

    Parameters
    ----------
    pipeline_module : PypelineModule
        Pipeline module.

    Returns
    -------
    NoneType
        None
    """

    input_ports = list(pipeline_module._m_input_ports.keys())

    if len(input_ports) == 1:
        input_shape = pipeline_module._m_input_ports[input_ports[0]].get_shape()
        print(f'Input port: {input_ports[0]} {input_shape}')

    else:
        print('Input ports:', end='')

        for i, item in enumerate(input_ports):
            input_shape = pipeline_module._m_input_ports[input_ports[i]].get_shape()

            if i < len(input_ports) - 1:
                print(f' {item} {input_shape},', end='')
            else:
                print(f' {item} {input_shape}')


# This function *cannot* be decorated with @typechecked, because the typeguard library checks type
# hints at *runtime*, when PypelineModule is not available without causing circular dependencies.
def output_info(pipeline_module: 'PypelineModule',
                output_shape: Dict[str, Tuple[int, ...]]) -> None:
    """
    Function to print information about the output data.

    Parameters
    ----------
    pipeline_module : PypelineModule
        Pipeline module.
    output_shape : dict
        Dictionary with the output dataset names and shapes.

    Returns
    -------
    NoneType
        None
    """

    output_ports = list(pipeline_module._m_output_ports.keys())

    if len(output_ports) == 1:
        if output_ports[0][:11] != 'fits_header':
            print(f'Output port: {output_ports[0]} {output_shape[output_ports[0]]}')

    else:
        print('Output ports:', end='')

        for i, item in enumerate(output_ports):
            if i < len(output_ports) - 1:
                print(f' {item} {output_shape[output_ports[i]]},', end='')
            else:
                print(f' {item} {output_shape[output_ports[i]]}')



"""
Functions for point source analysis.
"""

import math

from typing import Optional, Tuple

import numpy as np

from typeguard import typechecked
from scipy.stats import t
from scipy.ndimage import gaussian_filter
from skimage.feature import hessian_matrix
from photutils.aperture import aperture_photometry, CircularAperture

from pynpoint.util.image import shift_image, center_subpixel, pixel_distance, select_annulus, \
                                cartesian_to_polar, create_mask
from pynpoint.util.psf import pca_psf_subtraction
from pynpoint.util.residuals import combine_residuals


def compute_aperture_flux_elements(image: np.ndarray,
                                   x_pos: float,
                                   y_pos: float,
                                   size: float,
                                   ignore: bool):
    """
    Computes the average fluxes inside apertures with the same separation from the center.
    This function can be used to to estimate the residual flux of a planet at position
    (x_pos, y_pos) and the respective noise elements with same separation (see function false_alarm)
    It can also be used to compute the noise apertures is if no planet is present
    (needed for contrast curves).

    Parameters
    ----------
    image : numpy.ndarray
        The input image as a 2D numpy array. For example, this could be a residual frame returned by
        a :class:`.PcaPsfSubtractionModule`.
    x_pos : float
        The planet position (in pixels) along the horizontal axis. The pixel coordinates of the
        bottom-left corner of the image are (-0.5, -0.5). If no planet is present x_pos and y_pos
        determine the separation from the center.
    y_pos : float
        The planet position (pix) along the vertical axis. The pixel coordinates of the bottom-left
        corner of the image are (-0.5, -0.5). If no planet is present x_pos and y_pos
        determine the separation from the center.
    size : float
        The radius of the reference apertures (in pixels). Usually, this value is chosen close to
        one half of the typical FWHM of the PSF (0.514 lambda over D for a perfect Airy pattern; in
        practice, however, the FWHM is often larger than this).
    ignore : bool
        Whether or not to ignore the immediate neighboring apertures for the noise estimate. This is
        desirable in case there are "self-subtraction wings" left and right of the planet which
        would bias the estimation of the noise level at the separation of the planet if not ignored.

    Returns
    -------
    ap_phot :
        A list of aperture photometry values. If a planet was present ap_phot[0] contains the flux
        of the planet and ap_phot[1:] contains the noise. If not planet was present ap_phot[...]
        gives the aperture photometry of the noise elements.
    """

    # Compute the center of the current frame (with subpixel precision) and use it to compute the
    # radius of the given position in polar coordinates (with the origin at the center of the frame)
    center = center_subpixel(image)
    radius = math.sqrt((center[0] - y_pos)**2 + (center[1] - x_pos)**2)

    # Compute the number of apertures which we can place at the separation of  the given position
    num_ap = int(math.pi * radius / size)

    # Compute the angles at which to place the reference apertures
    ap_theta = np.linspace(0, 2 * math.pi, num_ap, endpoint=False)

    # If ignore is True, delete the apertures immediately right and left of the aperture placed on
    # the planet signal. These apertures often contain "self-subtraction wings", which means they
    # cannot be considered to originate from the same distribution. In accordance with section 3.2
    # of Mawet et al. (2014), such apertures are ignored to prevent bias.
    if ignore:
        num_ap -= 2
        ap_theta = np.delete(ap_theta, [1, np.size(ap_theta) - 1])

    # If the number of apertures is 2 or less, we cannot compute the false positive fraction
    if num_ap < 3:
        raise ValueError(
            f'Number of apertures (num_ap={num_ap}) is too small to calculate the '
            'false positive fraction.')

    # Initialize a numpy array in which we will store the integrated flux of all reference apertures
    ap_phot = np.zeros(num_ap)

    # Loop over all reference apertures and measure the integrated flux
    for i, theta in enumerate(ap_theta):
        # Compute the position of the current aperture in polar coordinates and convert to Cartesian
        x_tmp = center[1] + (x_pos - center[1]) * math.cos(theta) - \
            (y_pos - center[0]) * math.sin(theta)
        y_tmp = center[0] + (x_pos - center[1]) * math.sin(theta) + \
            (y_pos - center[0]) * math.cos(theta)

        # Place a circular aperture at a position and sum up the flux inside the aperture
        aperture = CircularAperture((x_tmp, y_tmp), size)
        phot_table = aperture_photometry(image, aperture, method='exact')

        ap_phot[i] = phot_table['aperture_sum']

    return ap_phot


@typechecked
def false_alarm(image: np.ndarray,
                x_pos: float,
                y_pos: float,
                size: float,
                ignore: bool) -> Tuple[float, float, float, float]:
    """
    Compute the signal-to-noise ratio (SNR), which is formally defined as the test statistic of a
    two-sample t-test, and related quantities (such as the FPF) at a given position in an image.

    For more detailed information about the definition of the signal-to-noise ratio and the
    motivation behind it, please see the following paper:

        Mawet, D. et al. (2014): "Fundamental limitations of high contrast imaging set by small
        sample statistics". *The Astrophysical Journal*, 792(2), 97.
        DOI: `10.1088/0004-637X/792/2/97 <https://dx.doi.org/10.1088/0004-637X/792/2/97>`_.

    Parameters
    ----------
    image : numpy.ndarray
        The input image as a 2D numpy array. For example, this could be a residual frame returned by
        a :class:`.PcaPsfSubtractionModule`.
    x_pos : float
        The planet position (in pixels) along the horizontal axis. The pixel coordinates of the
        bottom-left corner of the image are (-0.5, -0.5).
    y_pos : float
        The planet position (pix) along the vertical axis. The pixel coordinates of the bottom-left
        corner of the image are (-0.5, -0.5).
    size : float
        The radius of the reference apertures (in pixels). Usually, this value is chosen close to
        one half of the typical FWHM of the PSF (0.514 lambda over D for a perfect Airy pattern; in
        practice, however, the FWHM is often larger than this).
    ignore : bool
        Whether or not to ignore the immediate neighboring apertures for the noise estimate. This is
        desirable in case there are "self-subtraction wings" left and right of the planet which
        would bias the estimation of the noise level at the separation of the planet if not ignored.

    Returns
    -------
    signal_sum :
        The integrated (summed up) flux inside the signal aperture.

        Please note that this is **not** identical to the numerator of the fraction defining the SNR
        (which is given by the `signal_sum` minus the mean of the noise apertures).
    noise :
        The denominator of the SNR, i.e., the standard deviation of the integrated flux of the noise
        apertures, times a correction factor that accounts for small sample statistics.
    snr :
        The signal-to-noise ratio (SNR) as defined by Mawet et al. (2014) in eq. (8).
    fpf :
        The false positive fraction (FPF) as defined by Mawet et al. (2014) in eq. (10).
    """

    ap_phot = compute_aperture_flux_elements(image=image,
                                             x_pos=x_pos,
                                             y_pos=y_pos,
                                             size=size,
                                             ignore=ignore)

    # Define shortcuts to the signal and the noise aperture sums
    signal_aperture = ap_phot[0]
    noise_apertures = ap_phot[1:]

    # Compute the "signal", that is, the numerator of the signal-to-noise ratio: According to
    # eq. (8) in Mawet et al. (2014), this is given by the difference between the integrated flux
    # in the signal aperture and the mean of the integrated flux in the noise apertures
    signal = signal_aperture - np.mean(noise_apertures)

    # Compute the "noise", that is, the denominator of the signal-to-noise-ratio: According to
    # eq. (8) in Mawet et al. (2014), this is given by the standard deviation of the integrated flux
    # in the noise apertures times a correction factor to account for the small sample statistics.
    # NOTE: `ddof=1` is a necessary argument for np.std() in order to compute the *unbiased*
    #       estimate (i.e., including Bessel's corrections) of the standard deviation.
    noise = np.std(noise_apertures, ddof=1) *\
        math.sqrt(1 + 1 / (noise_apertures.shape[0]))

    # Compute the signal-to-noise ratio by dividing the "signal" through the "noise"
    snr = signal / noise

    # Compute the false positive fraction (FPF). According to eq. (10) in Mawet et al. (2014), the
    # FPF is given by 1 - F_nu(SNR), where F_nu is the cumulative distribution function (CDF) of a
    # t-distribution with `nu = n-1` degrees of freedom (see Section 3 of Mawet et al. (2014) for
    # more details on the Student's t distribution).
    # For numerical reasons, we use the survival function (SF), which is defined precisely as 1-CDF,
    # but may give more accurate results according to the scipy documentation.
    fpf = t.sf(snr, df=(noise_apertures.shape[0] - 1))

    return signal_aperture, noise, snr, fpf


@typechecked
def student_t(t_input: Tuple[str, float],
              radius: float,
              size: float,
              ignore: bool) -> float:
    """
    Function to calculate the false positive fraction for a given sigma level (Mawet et al. 2014).

    Parameters
    ----------
    t_input : tuple(str, float)
        Tuple with the input type ('sigma' or 'fpf') and the input value.
    radius : float
        Aperture radius (in pixels).
    size : float
        Separation of the aperture center from the center of the frame (in pixels).
    ignore : bool
        Whether or not to ignore the immediate neighboring apertures of the point source to exclude
        potential self-subtraction lobes.

    Returns
    -------
    float
        False positive fraction (FPF).
    """

    num_ap = int(math.pi * radius / size)

    if ignore:
        num_ap -= 2

    # Note that the number of degrees of freedom is given by nu = n-1 with n the number of samples.
    # The number of samples is equal to the number of apertures minus 1 (i.e. the planet aperture).
    # See Section 3 of Mawet et al. (2014) for more details on the Student's t distribution.

    if t_input[0] == 'sigma':
        t_result = t.sf(t_input[1], num_ap-2, loc=0., scale=1.)

    elif t_input[0] == 'fpf':
        t_result = t.ppf(1. - t_input[1], num_ap-2, loc=0., scale=1.)

    else:
        raise ValueError('First element of t_input needs to be "sigma" or "fpf"!')

    return t_result


@typechecked
def fake_planet(images: np.ndarray,
                psf: np.ndarray,
                parang: np.ndarray,
                position: Tuple[float, float],
                magnitude: float,
                psf_scaling: float,
                interpolation: str = 'spline') -> np.ndarray:
    """
    Function to inject artificial planets in a dataset.

    Parameters
    ----------
    images : numpy.ndarray
        Input images (3D).
    psf : numpy.ndarray
        PSF template (3D).
    parang : numpy.ndarray
        Parallactic angles (deg).
    position : tuple(float, float)
        Separation (pix) and position angle (deg) measured in counterclockwise with respect to the
        upward direction.
    magnitude : float
        Magnitude difference used to scale input PSF.
    psf_scaling : float
        Extra factor used to scale input PSF.
    interpolation : str
        Interpolation type ('spline', 'bilinear', or 'fft').

    Returns
    -------
    numpy.ndarray
        Images with artificial planet injected.
    """

    sep = position[0]
    ang = np.radians(position[1] + 90. - parang)

    flux_ratio = 10. ** (-magnitude / 2.5)
    psf = psf*psf_scaling*flux_ratio

    x_shift = sep*np.cos(ang)
    y_shift = sep*np.sin(ang)

    im_shift = np.zeros(images.shape)

    for i in range(images.shape[0]):
        if psf.shape[0] == 1:
            im_shift[i, ] = shift_image(psf[0, ],
                                        (float(y_shift[i]), float(x_shift[i])),
                                        interpolation,
                                        mode='reflect')

        else:
            im_shift[i, ] = shift_image(psf[i, ],
                                        (float(y_shift[i]), float(x_shift[i])),
                                        interpolation,
                                        mode='reflect')

    return images + im_shift


@typechecked
def merit_function(residuals: np.ndarray,
                   merit: str,
                   aperture: Tuple[int, int, float],
                   sigma: float,
                   var_noise: Optional[float]) -> float:
    """
    Function to calculate the figure of merit at a given position in the image residuals.

    Parameters
    ----------
    residuals : numpy.ndarray
        Residuals of the PSF subtraction (2D).
    merit : str
        Figure of merit for the chi-square function ('hessian', 'poisson', or 'gaussian').
    aperture : tuple(int, int, float)
        Position (y, x) of the aperture center (pix) and aperture radius (pix).
    sigma : float
        Standard deviation (pix) of the Gaussian kernel which is used to smooth the residuals
        before the chi-square is calculated.
    var_noise : float, None
        Variance of the noise which is required when `merit` is set to 'gaussian' or 'hessian'.

    Returns
    -------
    float
        Chi-square value.
    """

    rr_grid, _, _ = pixel_distance(residuals.shape, position=(aperture[0], aperture[1]))

    indices = np.where(rr_grid <= aperture[2])

    if merit == 'hessian':

        hessian_rr, hessian_rc, hessian_cc = hessian_matrix(image=residuals,
                                                            sigma=sigma,
                                                            mode='constant',
                                                            cval=0.,
                                                            order='rc',
                                                            use_gaussian_derivatives=False)

        hes_det = (hessian_rr*hessian_cc) - (hessian_rc*hessian_rc)

        chi_square = np.sum(hes_det[indices]**2)/var_noise

    elif merit == 'poisson':

        if sigma > 0.:
            residuals = gaussian_filter(input=residuals, sigma=sigma)

        chi_square = np.sum(np.abs(residuals[indices]))

    elif merit == 'gaussian':

        chi_square = np.sum(residuals[indices]**2)/var_noise

    else:

        raise ValueError('Figure of merit not recognized. Please use \'hessian\', \'poisson\' '
                         'or \'gaussian\'. Previous use of \'sum\' should now be set as '
                         '\'poisson\'.')

    return chi_square


@typechecked
def pixel_variance(var_type: str,
                   images: np.ndarray,
                   parang: np.ndarray,
                   cent_size: Optional[float],
                   edge_size: Optional[float],
                   pca_number: int,
                   residuals: str,
                   aperture: Tuple[int, int, float],
                   sigma: float) -> float:
    """
    Function to calculate the variance of the noise. After the PSF subtraction, images are rotated
    in opposite direction of the regular derotation, therefore dispersing any companion or disk
    signal. The noise is measured within an annulus.

    Parameters
    ----------
    var_type : str
        Variance type ('gaussian' or 'hessian').
    images : numpy.ndarray
        Input images (3D).
    parang : numpy.ndarray
        Parallactic angles.
    cent_size : float, None
        Radius of the central mask (pix). No mask is used when set to None.
    edge_size : float, None
        Outer radius (pix) beyond which pixels are masked. No outer mask is used when set to
        None.
    pca_number : int
        Number of principal components (PCs) used for the PSF subtraction.
    residuals : str
        Method for combining the residuals ('mean', 'median', 'weighted', or 'clipped').
    aperture : tuple(int, int, float)
        Aperture position (y, x) and radius (pix).
    sigma : float, None
        Standard deviation (pix) of the Gaussian kernel which is used to smooth the images.

    Returns
    -------
    float
        Variance of the pixel values. Either the variance of the pixel values ('gaussian') or
        the variance of the determinant of the Hessian ('hessian').
    """

    mask = create_mask(images.shape[-2:], (cent_size, edge_size))

    _, im_res_derot = pca_psf_subtraction(images*mask, parang, pca_number)

    res_noise = combine_residuals(residuals, im_res_derot)

    sep_ang = cartesian_to_polar(center_subpixel(res_noise), aperture[0], aperture[1])

    if var_type == 'gaussian':
        selected = select_annulus(res_noise[0, ], sep_ang[0]-aperture[2], sep_ang[0]+aperture[2])

    elif var_type == 'hessian':
        hessian_rr, hessian_rc, hessian_cc = hessian_matrix(image=res_noise[0, ],
                                                            sigma=sigma,
                                                            mode='constant',
                                                            cval=0.,
                                                            order='rc',
                                                            use_gaussian_derivatives=False)

        hes_det = (hessian_rr*hessian_cc) - (hessian_rc*hessian_rc)

        selected = select_annulus(hes_det, sep_ang[0]-aperture[2], sep_ang[0]+aperture[2])

    return float(np.var(selected))



"""
Functions for combining the residuals of the PSF subtraction.
"""

from typing import Optional

import numpy as np

from typeguard import typechecked
from scipy.ndimage import rotate


@typechecked
def combine_residuals(method: str,
                      res_rot: np.ndarray,
                      residuals: Optional[np.ndarray] = None,
                      angles: Optional[np.ndarray] = None) -> np.ndarray:
    """
    Wavelength wrapper for the combine_residual function. Produces an array with either 1
    or number of wavelengths sized array.

    Parameters
    ----------
    method : str
        Method used for combining the residuals ('mean', 'median', 'weighted', or 'clipped').
    res_rot : np.ndarray
        Derotated residuals of the PSF subtraction (3D).
    residuals : np.ndarray, None
        Non-derotated residuals of the PSF subtraction (3D). Only required for the noise-weighted
        residuals.
    angles : np.ndarray, None
        Derotation angles (deg). Only required for the noise-weighted residuals.

    Returns
    -------
    np.ndarray
        Collapsed residuals (3D).
    """

    if res_rot.ndim == 3:
        output = _residuals(method=method,
                            res_rot=np.asarray(res_rot),
                            residuals=residuals,
                            angles=angles)

    if res_rot.ndim == 4:
        output = np.zeros((res_rot.shape[0], res_rot.shape[2], res_rot.shape[3]))

        for i in range(res_rot.shape[0]):
            if residuals is None:
                output[i, ] = _residuals(method=method,
                                         res_rot=res_rot[i, ],
                                         residuals=residuals,
                                         angles=angles)[0]

            else:
                output[i, ] = _residuals(method=method,
                                         res_rot=res_rot[i, ],
                                         residuals=residuals[i, ],
                                         angles=angles)[0]

    return output


@typechecked
def _residuals(method: str,
               res_rot: np.ndarray,
               residuals: Optional[np.ndarray] = None,
               angles: Optional[np.ndarray] = None) -> np.ndarray:
    """
    Function for combining the derotated residuals of the PSF subtraction.

    Parameters
    ----------
    method : str
        Method used for combining the residuals ('mean', 'median', 'weighted', or 'clipped').
    res_rot : np.ndarray
        Derotated residuals of the PSF subtraction (3D).
    residuals : np.ndarray, None
        Non-derotated residuals of the PSF subtraction (3D). Only required for the noise-weighted
        residuals.
    angles : np.ndarray, None
        Derotation angles (deg). Only required for the noise-weighted residuals.

    Returns
    -------
    np.ndarray
        Combined residuals (3D).
    """

    if method == 'mean':
        stack = np.mean(res_rot, axis=0)

    elif method == 'median':
        stack = np.median(res_rot, axis=0)

    elif method == 'weighted':
        tmp_res_var = np.var(residuals, axis=0)

        res_repeat = np.repeat(tmp_res_var[np.newaxis, :, :],
                               repeats=residuals.shape[0],
                               axis=0)

        res_var = np.zeros(res_repeat.shape)

        for j, angle in enumerate(angles):
            # scipy.ndimage.rotate rotates in clockwise direction for positive angles
            res_var[j, ] = rotate(input=res_repeat[j, ],
                                  angle=angle,
                                  reshape=False)

        weight1 = np.divide(res_rot,
                            res_var,
                            out=np.zeros_like(res_var),
                            where=(np.abs(res_var) > 1e-100) & (res_var != np.nan))

        weight2 = np.divide(1.,
                            res_var,
                            out=np.zeros_like(res_var),
                            where=(np.abs(res_var) > 1e-100) & (res_var != np.nan))

        sum1 = np.sum(weight1, axis=0)
        sum2 = np.sum(weight2, axis=0)

        stack = np.divide(sum1,
                          sum2,
                          out=np.zeros_like(sum2),
                          where=(np.abs(sum2) > 1e-100) & (sum2 != np.nan))

    elif method == 'clipped':
        stack = np.zeros(res_rot.shape[-2:])

        for i in range(stack.shape[0]):
            for j in range(stack.shape[1]):
                pix_line = res_rot[:, i, j]

                if np.var(pix_line) > 0.:
                    no_mean = pix_line - np.mean(pix_line)

                    part1 = no_mean.compress((no_mean < 3.*np.sqrt(np.var(no_mean))).flat)
                    part2 = part1.compress((part1 > -3.*np.sqrt(np.var(no_mean))).flat)

                    stack[i, j] = np.mean(pix_line) + np.mean(part2)

    return stack[np.newaxis, ...]



import os

from urllib.request import urlretrieve

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.limits import ContrastCurveModule, MassLimitsModule
from pynpoint.processing.psfpreparation import AngleInterpolationModule
from pynpoint.util.tests import create_config, create_star_data, remove_test_data


class TestLimits:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_star_data(self.test_dir+'self.limits', npix=21, pos_star=10.)
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(path=self.test_dir,
                         folders=['self.limits'],
                         files=['model.AMES-Cond-2000.M-0.0.NaCo.Vega'])

    def test_read_data(self) -> None:

        module = FitsReadingModule(name_in='read',
                                   image_tag='read',
                                   input_dir=self.test_dir+'self.limits')

        self.pipeline.add_module(module)
        self.pipeline.run_module('read')

        data = self.pipeline.get_data('read')
        assert np.sum(data) == pytest.approx(108.43655133957289, rel=self.limit, abs=0.)
        assert data.shape == (10, 21, 21)

    def test_angle_interpolation(self) -> None:

        module = AngleInterpolationModule(name_in='angle',
                                          data_tag='read')

        self.pipeline.add_module(module)
        self.pipeline.run_module('angle')

        attr = self.pipeline.get_attribute('read', 'PARANG', static=False)
        assert np.sum(attr) == pytest.approx(900., rel=self.limit, abs=0.)
        assert attr.shape == (10, )

    def test_contrast_curve(self) -> None:

        proc = ['single', 'multi']

        for item in proc:

            if item == 'multi':
                with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'a') as hdf_file:
                    hdf_file['config'].attrs['CPU'] = 4

            module = ContrastCurveModule(name_in='contrast_'+item,
                                         image_in_tag='read',
                                         psf_in_tag='read',
                                         contrast_out_tag='limits_'+item,
                                         separation=(0.2, 0.3, 0.2),
                                         angle=(0., 360., 180.),
                                         threshold=('sigma', 5.),
                                         psf_scaling=1.,
                                         aperture=0.05,
                                         pca_number=2,
                                         cent_size=None,
                                         edge_size=1.,
                                         extra_rot=0.)

            self.pipeline.add_module(module)
            self.pipeline.run_module('contrast_'+item)

            data = self.pipeline.get_data('limits_'+item)
            assert data[0, 0] == pytest.approx(0.2, rel=self.limit, abs=0.)
            assert data[0, 1] == pytest.approx(2.580878183791224, rel=self.limit, abs=0.)
            assert data[0, 2] == pytest.approx(0.0007097688120261913, rel=self.limit, abs=0.)
            assert data[0, 3] == pytest.approx(0.00020126490906225968, rel=self.limit, abs=0.)
            assert data.shape == (1, 4)

    def test_contrast_curve_fpf(self) -> None:

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'a') as hdf_file:
            hdf_file['config'].attrs['CPU'] = 1

        module = ContrastCurveModule(name_in='contrast_fpf',
                                     image_in_tag='read',
                                     psf_in_tag='read',
                                     contrast_out_tag='limits_fpf',
                                     separation=(0.2, 0.3, 0.2),
                                     angle=(0., 360., 180.),
                                     threshold=('fpf', 1e-6),
                                     psf_scaling=1.,
                                     aperture=0.05,
                                     pca_number=2,
                                     cent_size=None,
                                     edge_size=1.,
                                     extra_rot=0.)

        self.pipeline.add_module(module)
        self.pipeline.run_module('contrast_fpf')

        data = self.pipeline.get_data('limits_fpf')
        assert data[0, 0] == pytest.approx(0.2, rel=self.limit, abs=0.)
        assert data[0, 1] == pytest.approx(1.9339430843041776, rel=self.limit, abs=0.)
        assert data[0, 2] == pytest.approx(0.000709768812026221, rel=self.limit, abs=0.)
        assert data[0, 3] == pytest.approx(1e-06, rel=self.limit, abs=0.)
        assert data.shape == (1, 4)

    def test_mass_limits(self) -> None:

        separation = np.linspace(0.1, 1.0, 10)
        contrast = -2.5*np.log10(1e-4/separation)
        variance = 0.1*contrast

        limits = np.zeros((10, 4))
        limits[:, 0] = separation
        limits[:, 1] = contrast
        limits[:, 2] = variance

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'a') as hdf_file:
            hdf_file['contrast_limits'] = limits

        url = 'https://home.strw.leidenuniv.nl/~stolker/pynpoint/' \
              'model.AMES-Cond-2000.M-0.0.NaCo.Vega'

        filename = self.test_dir + 'model.AMES-Cond-2000.M-0.0.NaCo.Vega'

        urlretrieve(url, filename)

        module = MassLimitsModule(model_file=filename,
                                  star_prop={'magnitude': 10., 'distance': 100., 'age': 20.},
                                  name_in='mass',
                                  contrast_in_tag='contrast_limits',
                                  mass_out_tag='mass_limits',
                                  instr_filter='L\'')

        self.pipeline.add_module(module)
        self.pipeline.run_module('mass')

        data = self.pipeline.get_data('mass_limits')
        assert np.mean(data[:, 0]) == pytest.approx(0.55, rel=self.limit, abs=0.)
        assert np.mean(data[:, 1]) == pytest.approx(0.001891690765603738, rel=self.limit, abs=0.)
        assert np.mean(data[:, 2]) == pytest.approx(0.000964309686441908, rel=self.limit, abs=0.)
        assert np.mean(data[:, 3]) == pytest.approx(-0.000696402843279597, rel=self.limit, abs=0.)
        assert data.shape == (10, 4)



import os

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.resizing import CropImagesModule, ScaleImagesModule, \
                                         AddLinesModule, RemoveLinesModule
from pynpoint.util.tests import create_config, create_star_data, create_ifs_data, remove_test_data


class TestResizing:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_star_data(self.test_dir+'resize')
        create_ifs_data(self.test_dir+'resize_ifs')
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=['resize', 'resize_ifs'])

    def test_read_data(self) -> None:

        module = FitsReadingModule(name_in='read',
                                   image_tag='read',
                                   input_dir=self.test_dir+'resize',
                                   overwrite=True,
                                   check=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('read')

        data = self.pipeline.get_data('read')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

        module = FitsReadingModule(name_in='read_ifs',
                                   image_tag='read_ifs',
                                   input_dir=self.test_dir+'resize_ifs',
                                   overwrite=True,
                                   check=True,
                                   ifs_data=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('read_ifs')

        data = self.pipeline.get_data('read_ifs')
        assert np.sum(data) == pytest.approx(749.8396528807369, rel=self.limit, abs=0.)
        assert data.shape == (3, 10, 21, 21)

    def test_crop_images(self) -> None:

        module = CropImagesModule(size=0.2,
                                  center=None,
                                  name_in='crop1',
                                  image_in_tag='read',
                                  image_out_tag='crop1')

        self.pipeline.add_module(module)
        self.pipeline.run_module('crop1')

        module = CropImagesModule(size=0.2,
                                  center=(4, 4),
                                  name_in='crop2',
                                  image_in_tag='read',
                                  image_out_tag='crop2')

        self.pipeline.add_module(module)
        self.pipeline.run_module('crop2')

        module = CropImagesModule(size=0.2,
                                  center=(4, 4),
                                  name_in='crop_ifs',
                                  image_in_tag='read_ifs',
                                  image_out_tag='crop_ifs')

        self.pipeline.add_module(module)
        self.pipeline.run_module('crop_ifs')

        data = self.pipeline.get_data('crop1')
        assert np.sum(data) == pytest.approx(104.93318507061295, rel=self.limit, abs=0.)
        assert data.shape == (10, 9, 9)

        data = self.pipeline.get_data('crop2')
        assert np.sum(data) == pytest.approx(105.64863165433025, rel=self.limit, abs=0.)
        assert data.shape == (10, 9, 9)

        data = self.pipeline.get_data('crop_ifs')
        assert np.sum(data) == pytest.approx(15.870936600122521, rel=self.limit, abs=0.)
        assert data.shape == (3, 10, 9, 9)

    def test_scale_images(self) -> None:

        module = ScaleImagesModule(name_in='scale1',
                                   image_in_tag='read',
                                   image_out_tag='scale1',
                                   scaling=(2., 2., None),
                                   pixscale=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('scale1')

        module = ScaleImagesModule(name_in='scale2',
                                   image_in_tag='read',
                                   image_out_tag='scale2',
                                   scaling=(None, None, 2.),
                                   pixscale=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('scale2')

        data = self.pipeline.get_data('scale1')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 22, 22)

        data = self.pipeline.get_data('scale2')
        assert np.sum(data) == pytest.approx(211.08557759610554, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

        attr = self.pipeline.get_attribute('read', 'PIXSCALE', static=True)
        assert attr == pytest.approx(0.027, rel=self.limit, abs=0.)

        attr = self.pipeline.get_attribute('scale1', 'PIXSCALE', static=True)
        assert attr == pytest.approx(0.0135, rel=self.limit, abs=0.)

        attr = self.pipeline.get_attribute('scale2', 'PIXSCALE', static=True)
        assert attr == pytest.approx(0.027, rel=self.limit, abs=0.)

    def test_add_lines(self) -> None:

        module = AddLinesModule(lines=(2, 5, 0, 3),
                                name_in='add',
                                image_in_tag='read',
                                image_out_tag='add')

        self.pipeline.add_module(module)
        self.pipeline.run_module('add')

        data = self.pipeline.get_data('add')
        assert np.sum(data) == pytest.approx(105.54278879805275, rel=self.limit, abs=0.)
        assert data.shape == (10, 14, 18)

    def test_remove_lines(self) -> None:

        module = RemoveLinesModule(lines=(2, 5, 0, 3),
                                   name_in='remove',
                                   image_in_tag='read',
                                   image_out_tag='remove')

        self.pipeline.add_module(module)
        self.pipeline.run_module('remove')

        data = self.pipeline.get_data('remove')
        assert np.sum(data) == pytest.approx(67.49726677462391, rel=self.limit, abs=0.)
        assert data.shape == (10, 8, 4)

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'a') as hdf_file:
            hdf_file['config'].attrs['CPU'] = 4

        self.pipeline.run_module('remove')

        data_multi = self.pipeline.get_data('remove')
        assert data == pytest.approx(data_multi, rel=self.limit, abs=0.)
        assert data.shape == data_multi.shape



import os
import h5py

import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.psfsubtraction import PcaPsfSubtractionModule
from pynpoint.util.tests import create_config, create_ifs_data, remove_test_data


class TestPsfSubtractionSdi:

    def setup_class(self) -> None:

        self.limit = 1e-5
        self.test_dir = os.path.dirname(__file__) + '/'

        create_ifs_data(self.test_dir+'science')
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=['science'])

    def test_read_data(self) -> None:

        module = FitsReadingModule(name_in='read',
                                   image_tag='science',
                                   input_dir=self.test_dir+'science',
                                   ifs_data=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('read')

        data = self.pipeline.get_data('science')
        assert np.sum(data) == pytest.approx(749.8396528807368, rel=self.limit, abs=0.)
        assert data.shape == (3, 10, 21, 21)

        self.pipeline.set_attribute('science', 'WAVELENGTH', [1., 1.1, 1.2], static=False)
        self.pipeline.set_attribute('science', 'PARANG', np.linspace(0., 180., 10), static=False)

    def test_psf_subtraction_sdi(self) -> None:

        processing_types = ['ADI', 'SDI+ADI', 'ADI+SDI', 'CODI']

        expected = [[-0.16718942968552664, -0.790697125718532,
                     19.507979777136892, -0.21617058715490922],
                    [-0.001347198747121658, -0.08621264803633322,
                     2.3073192270025333, -0.010269745733878437],
                    [0.009450917836998779, -0.05776205365084376,
                     -0.43506678222476264, 0.0058856438951644455],
                    [-0.2428739554898396, -0.5069023645693083,
                     9.326414176548905, 0.00]]

        shape_expc = [(2, 3, 21, 21), (2, 2, 3, 21, 21), (1, 1, 3, 21, 21), (2, 3, 21, 21)]

        pca_numbers = [range(1, 3), (range(1, 3), range(1, 3)), ([1], [1]), range(1, 3)]

        res_arr_tags = [None, None, 'res_arr_single_sdi_ADI+SDI', None]

        for i, p_type in enumerate(processing_types):

            module = PcaPsfSubtractionModule(pca_numbers=pca_numbers[i],
                                             name_in='pca_single_sdi_'+p_type,
                                             images_in_tag='science',
                                             reference_in_tag='science',
                                             res_mean_tag='res_mean_single_sdi_'+p_type,
                                             res_median_tag='res_median_single_sdi_'+p_type,
                                             res_weighted_tag='res_weighted_single_sdi_'+p_type,
                                             res_rot_mean_clip_tag='res_clip_single_sdi_'+p_type,
                                             res_arr_out_tag=res_arr_tags[i],
                                             basis_out_tag='basis_single_sdi_'+p_type,
                                             extra_rot=0.,
                                             subtract_mean=True,
                                             processing_type=p_type)

            self.pipeline.add_module(module)
            self.pipeline.run_module('pca_single_sdi_'+p_type)

            data = self.pipeline.get_data('res_mean_single_sdi_'+p_type)
            assert np.sum(data) == pytest.approx(expected[i][0], rel=self.limit, abs=0.)
            assert data.shape == shape_expc[i]

            data = self.pipeline.get_data('res_median_single_sdi_'+p_type)
            assert np.sum(data) == pytest.approx(expected[i][1], rel=self.limit, abs=0.)
            assert data.shape == shape_expc[i]

            data = self.pipeline.get_data('res_weighted_single_sdi_'+p_type)
            assert np.sum(data) == pytest.approx(expected[i][2], rel=self.limit, abs=0.)
            assert data.shape == shape_expc[i]

            data = self.pipeline.get_data('res_clip_single_sdi_'+p_type)
#            assert np.sum(data) == pytest.approx(expected[i][3], rel=self.limit, abs=0.)
            assert data.shape == shape_expc[i]

            # data = self.pipeline.get_data('basis_single_sdi_'+p_type)
            # assert np.sum(data) == pytest.approx(-1.3886119555248766, rel=self.limit, abs=0.)
            # assert data.shape == (5, 30, 30)

    def test_psf_subtraction_sdi_multi(self) -> None:

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'a') as hdf_file:
            hdf_file['config'].attrs['CPU'] = 4

        processing_types = ['SDI', 'ADI+SDI']

        pca_numbers = [range(1, 3), (range(1, 3), range(1, 3))]

        expected = [[-0.004159475403024583, 0.02613693149969979,
                     -0.12940723035023394, -0.008432530081399985],
                    [-0.006580571531064533, -0.08171546066331437,
                     0.5700432018961117, -0.014527353460544753]]

        shape_expc = [(2, 3, 21, 21), (2, 2, 3, 21, 21)]

        for i, p_type in enumerate(processing_types):

            module = PcaPsfSubtractionModule(pca_numbers=pca_numbers[i],
                                             name_in='pca_multi_sdi_'+p_type,
                                             images_in_tag='science',
                                             reference_in_tag='science',
                                             res_mean_tag='res_mean_multi_sdi_'+p_type,
                                             res_median_tag='res_median_multi_sdi_'+p_type,
                                             res_weighted_tag='res_weighted_multi_sdi_'+p_type,
                                             res_rot_mean_clip_tag='res_clip_multi_sdi_'+p_type,
                                             res_arr_out_tag=None,
                                             basis_out_tag=None,
                                             extra_rot=0.,
                                             subtract_mean=True,
                                             processing_type=p_type)

            self.pipeline.add_module(module)
            self.pipeline.run_module('pca_multi_sdi_'+p_type)

            data = self.pipeline.get_data('res_mean_multi_sdi_'+p_type)
            assert np.sum(data) == pytest.approx(expected[i][0], rel=self.limit, abs=0.)
            assert data.shape == shape_expc[i]

            data = self.pipeline.get_data('res_median_multi_sdi_'+p_type)
            assert np.sum(data) == pytest.approx(expected[i][1], rel=self.limit, abs=0.)
            assert data.shape == shape_expc[i]

            data = self.pipeline.get_data('res_weighted_multi_sdi_'+p_type)
            assert np.sum(data) == pytest.approx(expected[i][2], rel=self.limit, abs=0.)
            assert data.shape == shape_expc[i]

            data = self.pipeline.get_data('res_clip_multi_sdi_'+p_type)
#            assert np.sum(data) == pytest.approx(expected[i][3], rel=self.limit, abs=0.)
            assert data.shape == shape_expc[i]
